# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea test' command."""
import contextlib
import os
import sys
import tempfile
from pathlib import Path
from typing import Any, Callable, List, Optional, Sequence, Set, Tuple, Union, cast

import click
import pytest
from coverage.cmdline import main as coverage

from aea.aea_builder import AEABuilder
from aea.cli.utils.click_utils import (
    PublicIdParameter,
    determine_package_type_for_directory,
)
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project
from aea.cli.utils.package_utils import get_package_path
from aea.components.base import load_aea_package, perform_load_aea_package
from aea.configurations.base import (
    AgentConfig,
    ComponentConfiguration,
    PackageConfiguration,
)
from aea.configurations.constants import (
    AEA_TEST_DIRNAME,
    CONNECTION,
    CONTRACT,
    PROTOCOL,
    SKILL,
)
from aea.configurations.data_types import (
    ComponentId,
    ComponentType,
    PackageType,
    PublicId,
)
from aea.configurations.loader import (
    load_component_configuration,
    load_package_configuration,
)
from aea.configurations.manager import find_component_directory_from_component_id
from aea.exceptions import enforce
from aea.package_manager.v1 import PackageManagerV1


COVERAGERC_FILE = ".coveragerc"
COVERAGERC_CONFIG = """[run]
omit =
    */tests/*
    */.tox/*
    */*_pb2.py
    */*_pb2_*.py

[html]
directory = {root_dir}/htmlcov

[xml]
output = {root_dir}/coverage.xml
"""


@click.group(
    invoke_without_command=True,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.pass_context
@click.option(
    "--cov",
    is_flag=True,
    default=False,
    help="Use this flag to enable code coverage checks.",
)
@click.option(
    "--cov-output",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="Directory to output codecov reports.",
)
@click.option(
    "--append",
    help="Directory to output codecov reports.",
    is_flag=True,
)
def test(
    click_context: click.Context, cov: bool, cov_output: str, append: bool
) -> None:
    """Run tests of an AEA project."""
    ctx = cast(Context, click_context.obj)
    ctx.config["cov"] = cov
    ctx.config["cov_output"] = cov_output
    ctx.config["append"] = append

    if click_context.invoked_subcommand is None:
        test_aea_project(click_context, Path(ctx.cwd), args=[])


@check_aea_project
def test_aea_project(
    click_context: click.Context, aea_project_dirpath: Path, args: Sequence[str]
) -> None:
    """Run tests of an AEA project."""
    click.echo("Executing tests of the AEA project...")
    ctx = cast(Context, click_context.obj)
    test_package_by_path(
        aea_project_dirpath,
        args,
        aea_project_path=Path(ctx.cwd),
        skip_consistency_check=ctx.config.get("skip_consistency_check", False),
    )


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("connection_public_id", type=PublicIdParameter(), required=True)
@click.pass_context
def connection(
    click_context: click.Context,
    connection_public_id: PublicId,
) -> None:
    """Executes a test suite of a connection package dependency."""
    test_item(click_context.obj, CONNECTION, connection_public_id, click_context.args)


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("contract_public_id", type=PublicIdParameter(), required=True)
@click.pass_context
def contract(
    click_context: click.Context,
    contract_public_id: PublicId,
) -> None:
    """Executes a test suite of a contract package dependency."""
    test_item(click_context.obj, CONTRACT, contract_public_id, click_context.args)


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("protocol_public_id", type=PublicIdParameter(), required=True)
@click.pass_context
def protocol(
    click_context: click.Context,
    protocol_public_id: PublicId,
) -> None:
    """Executes a test suite of a protocol package dependency."""
    test_item(click_context.obj, PROTOCOL, protocol_public_id, click_context.args)


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("skill_public_id", type=PublicIdParameter(), required=True)
@click.pass_context
def skill(
    click_context: click.Context,
    skill_public_id: PublicId,
) -> None:
    """Executes a test suite of a skill package dependency."""
    test_item(click_context.obj, SKILL, skill_public_id, click_context.args)


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument(
    "path", type=click.Path(exists=True, file_okay=False, dir_okay=True), required=True
)
@click.pass_context
def by_path(
    click_context: click.Context,
    path: str,
) -> None:
    """Executes a test suite of a package specified by a path."""
    click.echo(f"Executing tests of package at {path}'...")

    ctx: Context = click_context.obj
    full_path = Path(ctx.cwd) / Path(path)

    test_package_by_path(
        full_path,
        click_context.args,
        packages_dir=Path(ctx.registry_path),
        cov=ctx.config.get("cov", False),
        cov_output=ctx.config.get("cov_output"),
        append_coverage=ctx.config.get("append", False),
        skip_consistency_check=ctx.config.get("skip_consistency_check", False),
    )


@test.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.option(
    "--author",
    "-a",
    type=str,
    multiple=True,
    help="Author name(s) to restrict tests to.",
)
@click.option(
    "--all",
    "all_",
    is_flag=True,
    show_default=True,
    default=False,
    help="Run test for all packages. By default dev only.",
)
@click.pass_context
def packages(click_context: click.Context, author: Tuple[str], all_: bool) -> None:
    """Executes a test suite for a collection of packages."""
    ctx: Context = click_context.obj
    packages_dir = Path(ctx.registry_path)
    available_packages = get_packages_list(
        packages_dir=packages_dir, packages_filter="all" if all_ else "dev"
    )

    cov = ctx.config.get("cov", False)
    cov_output = Path(ctx.config.get("cov_output") or Path.cwd()).resolve()

    with CoverageContext(root_dir=cov_output, append=True) as coverage_context:
        failures = test_package_collection(
            available_packages=available_packages,
            packages_dir=packages_dir,
            pytest_args=click_context.args,
            cov=cov,
            coverage_context=coverage_context,
            authors=author,
        )
        if cov:
            coverage_context.generate()

    if len(failures) > 0:
        click.echo("Failed tests")
        click.echo("Exit Code\tPackage")
        for exit_code, package in failures:
            click.echo(f"{exit_code}       \t{package}")
        sys.exit(1)


def get_packages_list(
    packages_dir: Path, packages_filter: str = "dev"
) -> List[Tuple[str, Path]]:
    """
    Get list of packages to test.

    :param packages_dir: Path
    :param packages_filter: str, ["all", "dev"]

    :return: List of tuples of pacakge type and path
    """
    package_manager = PackageManagerV1.from_dir(packages_dir)
    if packages_filter == "all":
        packages_ids = package_manager.all_packages.keys()
    elif packages_filter == "dev":
        packages_ids = package_manager.dev_packages.keys()
    else:
        raise ValueError(
            f"Unknown package filter: {packages_filter}. Valid are all,dev"
        )

    packages_list = [
        (
            str(package_id.package_type.to_plural()),
            package_manager.package_path_from_package_id(package_id),
        )
        for package_id in packages_ids
    ]
    return packages_list


class CoverageContext:
    """Coveragerc file context"""

    def __init__(
        self,
        root_dir: Path,
        append: bool = False,
    ) -> None:
        """Initialize object."""

        self.root_dir = root_dir
        self.append = append

        self._t = tempfile.TemporaryDirectory()
        self.coveragerc_file = Path(self._t.name, COVERAGERC_FILE)

    def pytest_args(
        self,
        test_path: Path,
    ) -> List[str]:
        """Returns coverage flags for pytest command"""

        args = [
            f"--cov-config={self.coveragerc_file}",
            "--cov-report=term",
            f"--cov={test_path}",
        ]

        if self.append:
            args.append("--cov-append")
        return args

    def __enter__(
        self,
    ) -> "CoverageContext":
        """Enter context."""
        self.coveragerc_file.write_text(
            COVERAGERC_CONFIG.format(root_dir=self.root_dir),
        )
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        """Exit context."""

        with contextlib.suppress(OSError, PermissionError):
            self._t.cleanup()

    def generate(
        self,
    ) -> None:
        """Generate coverage report."""

        coverage(argv=["html", f"--rcfile={self.coveragerc_file}"])
        coverage(argv=["xml", f"--rcfile={self.coveragerc_file}"])


def test_item(
    ctx: Context,
    item_type: str,
    item_public_id: PublicId,
    pytest_arguments: Sequence[str],
) -> None:
    """
    Run tests of a package dependency.

    :param ctx: the context.
    :param item_type: the item type.
    :param item_public_id: the item public id.
    :param pytest_arguments: arguments to forward to Pytest
    """
    click.echo(
        "Executing tests of component of type {}, {}' ...".format(
            item_type, item_public_id
        )
    )
    package_dirpath = Path(
        get_package_path(ctx.cwd, item_type, item_public_id, is_vendor=False)
    )
    if not package_dirpath.exists():
        # check if it is a vendor package
        package_dirpath = Path(
            get_package_path(ctx.cwd, item_type, item_public_id, is_vendor=True)
        )
        enforce(
            package_dirpath.exists(),
            exception_text=f"package {item_public_id} of type {item_type} not found",
            exception_class=click.ClickException,
        )
    # for a package in an AEA project, the "packages" dir is the AEA project dir
    aea_project_path = Path(ctx.cwd)
    test_package_by_path(
        package_dirpath,
        pytest_arguments,
        aea_project_path=aea_project_path,
        cov=ctx.config.get("cov", False),
        skip_consistency_check=ctx.config.get("skip_consistency_check", False),
    )


def load_package(
    package_dir: Path,
    aea_project_path: Optional[Path] = None,
    packages_dir: Optional[Path] = None,
    skip_consistency_check: bool = False,
) -> None:
    """Load packages into memory."""
    enforce(
        (aea_project_path is None) != (packages_dir is None),
        "one of either aea_project_path or packages_dir must be specified",
    )
    root_packages = aea_project_path if aea_project_path else packages_dir

    package_path_finder = (
        find_component_directory_from_component_id
        if aea_project_path
        else find_component_directory_from_component_id_in_registry
    )
    if root_packages is None:
        raise ValueError("Packages dir not set!")
    # check the path points to a valid AEA package
    package_type = determine_package_type_for_directory(package_dir)
    if package_type != PackageType.AGENT:
        component_type = ComponentType(package_type.value)
        configuration = load_component_configuration(
            component_type, package_dir, skip_consistency_check=skip_consistency_check
        )
        configuration.directory = package_dir
        load_aea_packages_recursively(configuration, package_path_finder, root_packages)

        test_package_dir = package_dir / AEA_TEST_DIRNAME
        enforce(
            test_package_dir.exists(),
            f"tests directory in {package_dir} not found",
            click.ClickException,
        )

    else:
        if aea_project_path:
            # for agent's workdir
            agent_config = AEABuilder.try_to_load_agent_configuration_file(
                cast(Path, aea_project_path)
            )
            load_aea_packages_recursively(
                agent_config, package_path_finder, root_packages
            )
        else:
            # for agents package in packages
            package_configuration = load_package_configuration(
                package_type=package_type,
                directory=package_dir,
                skip_aea_validation=skip_consistency_check,
            )
            package_configuration.directory = package_dir

            # loads dependencies packages
            load_aea_packages_recursively(
                package_configuration, package_path_finder, root_packages
            )

            # requires to populate sys.modules with packages/author/agents
            perform_load_aea_package(
                package_dir,
                package_configuration.author,
                package_type.to_plural(),
                package_configuration.name,
            )

            test_package_dir = package_dir / AEA_TEST_DIRNAME
            enforce(
                test_package_dir.exists(),
                f"tests directory in {package_dir} not found",
                click.ClickException,
            )


def test_package_by_path(
    package_dir: Path,
    pytest_arguments: Sequence[str],
    aea_project_path: Optional[Path] = None,
    packages_dir: Optional[Path] = None,
    cov: bool = False,
    cov_output: Optional[Path] = None,
    append_coverage: bool = False,
    skip_consistency_check: bool = False,
) -> None:
    """
    Test package placed in package_dir.

    :param package_dir: directory of the package
    :param pytest_arguments: arguments to forward to Pytest
    :param aea_project_path: directory to the AEA project
    :param packages_dir: directory of the packages to import from
    :param cov: coverage capture indicator
    :param cov_output: Path to coverage output directory
    :param append_coverage: Append test coverage instead of starting clean
    :param skip_consistency_check: skip the package consistency check
    """
    cov_output = Path(cov_output or Path.cwd()).resolve()

    load_package(
        package_dir,
        aea_project_path,
        packages_dir,
        skip_consistency_check=skip_consistency_check,
    )
    with CoverageContext(
        root_dir=cov_output, append=append_coverage
    ) as coverage_context:
        runtime_args = [
            *get_pytest_args(
                package_dir=package_dir,
                cov=cov,
                coverage_context=coverage_context,
            ),
            *pytest_arguments,
            "-vvvvvvvvvvvvvvv",
        ]
        exit_code = pytest.main(runtime_args)
        if cov:
            coverage_context.generate()

    sys.exit(exit_code)


def test_package_collection(
    available_packages: List[Tuple[str, Path]],
    packages_dir: Path,
    pytest_args: List[str],
    cov: bool,
    coverage_context: CoverageContext,
    authors: Tuple[str, ...],
) -> List[Tuple[int, str]]:
    """Test a collection of packages."""

    failures = []

    for package_type, package_dir in available_packages:
        if authors != () and not any(
            [author in str(package_dir) for author in authors]
        ):
            continue
        test_dir = package_dir / AEA_TEST_DIRNAME
        if not test_dir.exists():
            continue

        load_package(
            package_dir, packages_dir=packages_dir, skip_consistency_check=True
        )
        click.echo(f"Running tests for {package_dir.name} of type {package_type}")
        exit_code = pytest.main(
            [
                *get_pytest_args(
                    package_dir=package_dir, cov=cov, coverage_context=coverage_context
                ),
                *pytest_args,
            ]
        )

        if exit_code:
            if exit_code == pytest.ExitCode.NO_TESTS_COLLECTED:
                click.echo(
                    f"Could not collect tests for for {package_dir.name} of type {package_type}"
                )
                continue

            click.echo(
                f"Running tests for for {package_dir.name} of type {package_type} failed"
            )
            failures.append((exit_code, os.path.sep.join(package_dir.parts[-3:])))

    return failures


def load_aea_packages_recursively(
    config: Union[ComponentConfiguration, AgentConfig, PackageConfiguration],
    package_path_finder: Callable[[Path, ComponentId], Path],
    root_packages: Path,
    already_loaded: Optional[Set[ComponentId]] = None,
) -> None:
    """
    Load all AEA packages recursively.

    It works like 'load_aea_package', but recursively imports all dependencies.

    :param config: the component configuration
    :param package_path_finder: a function that find packages from the packages dir
    :param root_packages: the path to the root of packages dir
    :param already_loaded: the already loaded component ids
    """
    already_loaded = already_loaded if already_loaded else set()
    for dependency_id in config.package_dependencies:
        # TODO: load packages in topological order? Should not matter as at the moment we are not  # pylint: disable=fixme
        #       actually running the modules, just populating sys.modules
        dependency_path = package_path_finder(root_packages, dependency_id)
        dependency_configuration = load_component_configuration(
            dependency_id.component_type, dependency_path
        )
        dependency_configuration.directory = dependency_path
        load_aea_packages_recursively(
            dependency_configuration, package_path_finder, root_packages, already_loaded
        )
    if isinstance(config, ComponentConfiguration):
        # works for component, not for agent config or package config used for agent
        load_aea_package(config)
        already_loaded.add(config.component_id)


def find_component_directory_from_component_id_in_registry(
    registry_path: Path, component_id: ComponentId
) -> Path:
    """Find a component directory from component id in a registry."""
    package_path = (
        registry_path
        / component_id.author
        / component_id.component_type.to_plural()
        / component_id.public_id.name
    )
    if package_path.exists() and package_path.is_dir():
        return package_path

    raise ValueError("Package {} not found.".format(component_id))


def get_pytest_args(
    package_dir: Path,
    cov: bool = False,
    coverage_context: Optional[CoverageContext] = None,
) -> List:
    """Get pytest args for coverage checks."""
    args = [
        str(package_dir / AEA_TEST_DIRNAME),
    ]

    if cov:
        args.extend(
            cast(CoverageContext, coverage_context).pytest_args(test_path=package_dir)
        )

    return args



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test P2PLibp2p connection build."""

import os
import tempfile
import time
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture  # type: ignore

from aea.exceptions import AEAException

from packages.valory.connections.p2p_libp2p import check_dependencies
from packages.valory.connections.p2p_libp2p.check_dependencies import (
    LIBP2P_NODE_MODULE_NAME,
    MINIMUM_GCC_VERSION,
    MINIMUM_GO_VERSION,
    build_node,
)
from packages.valory.connections.p2p_libp2p.check_dependencies import (
    check_versions as base_check_versions,
)
from packages.valory.connections.p2p_libp2p.check_dependencies import version_to_string
from packages.valory.connections.p2p_libp2p.tests.base import SKIP_WINDOWS


def check_versions() -> None:
    """Otherwise buffer may occasionally be found empty"""
    base_check_versions()
    time.sleep(0.1)


def test_check_versions(capsys: CaptureFixture) -> None:
    """Test check_versions - positive case."""
    check_versions()
    cap = capsys.readouterr()
    assert f"check 'go'>={version_to_string(MINIMUM_GO_VERSION)}, found " in cap.out
    assert f"check 'gcc'>={version_to_string(MINIMUM_GCC_VERSION)}, found " in cap.out


def test_check_versions_negative_binary_not_found() -> None:
    """Test check_versions - negative case, binary not found."""
    with mock.patch("shutil.which", return_value=None):
        with pytest.raises(
            AEAException,
            match="'go' is required by the libp2p connection, but it is not installed, or it is not accessible from the system path.",
        ):
            check_versions()


def test_check_versions_negative_version_too_low() -> None:
    """Test check_versions - negative case, version too low."""
    with mock.patch.object(
        check_dependencies,
        "get_version",
        return_value=(0, 0, 0),
    ):
        with pytest.raises(
            AEAException,
            match=f"The installed version of 'go' is too low: expected at least {version_to_string(MINIMUM_GO_VERSION)}; found 0.0.0.",
        ):
            check_versions()


def test_check_versions_negative_cannot_parse_version(capsys: CaptureFixture) -> None:
    """Test the check_versions - negative case, cannot parse version."""

    with mock.patch("subprocess.check_output", return_value=b""):
        check_versions()
    cap = capsys.readouterr()
    assert (
        "Warning: cannot parse 'go' version from command: ['go', 'version']." in cap.out
    )
    assert (
        "Warning: cannot parse 'gcc' version from command: ['gcc', '--version']."
        in cap.out
    )


@SKIP_WINDOWS
def test_build_node() -> None:
    """Test build node function."""
    with tempfile.TemporaryDirectory() as build_dir:
        build_node(build_dir)
        assert os.path.exists(os.path.join(build_dir, LIBP2P_NODE_MODULE_NAME))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the http_echo skill."""
# pylint: skip-file

from pathlib import Path
from typing import cast

from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.skills.http_echo.dialogues import (
    DefaultDialogue,
    DefaultDialogues,
    HttpDialogue,
    HttpDialogues,
)
from packages.valory.protocols.http.message import HttpMessage


PACKAGE_DIR = Path(__file__).parent.parent


class TestDialogues(BaseSkillTestCase):
    """Test dialogue class of http_echo."""

    path_to_skill = PACKAGE_DIR

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.http_dialogues = cast(
            HttpDialogues, cls._skill.skill_context.http_dialogues
        )

    def test_default_dialogues(self):
        """Test the DefaultDialogues class."""
        _, dialogue = self.default_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )
        assert dialogue.role == DefaultDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_http_dialogues(self):
        """Test the HttpDialogues class."""
        _, dialogue = self.http_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=HttpMessage.Performative.REQUEST,
            method="some_method",
            url="some_url",
            version="some_version",
            headers="some_headers",
            body=b"some_body",
        )
        assert dialogue.role == HttpDialogue.Role.SERVER
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of AEA multiple instances runner."""
import asyncio
import logging
from asyncio import iscoroutine
from asyncio.events import AbstractEventLoop
from typing import Dict, Sequence, Type, cast

from aea.aea import AEA
from aea.helpers.multiple_executor import (
    AbstractExecutorTask,
    AbstractMultipleExecutor,
    AbstractMultipleRunner,
    AsyncExecutor,
    ExecutorExceptionPolicies,
    TaskAwaitable,
    ThreadExecutor,
)
from aea.runtime import AsyncRuntime


_default_logger = logging.getLogger(__name__)


class AEAInstanceTask(AbstractExecutorTask):
    """Task to run agent instance."""

    def __init__(self, agent: AEA) -> None:
        """
        Init AEA instance task.

        :param agent: AEA instance to run within task.
        """
        self._agent = agent
        super().__init__()

    @property
    def id(self) -> str:
        """Return agent name."""
        return self._agent.name

    def start(self) -> None:  # type: ignore
        """Start task."""
        try:
            self._agent.start()
        except BaseException:
            _default_logger.exception("Exceptions raised in runner task.")
            raise

    def stop(self) -> None:
        """Stop task."""
        self._agent.runtime.stop()

    def create_async_task(self, loop: AbstractEventLoop) -> TaskAwaitable:
        """
        Return asyncio Task for task run in asyncio loop.

        :param loop: abstract event loop
        :return: task to run runtime
        """
        self._agent.runtime.set_loop(loop)
        if not isinstance(self._agent.runtime, AsyncRuntime):  # pragma: nocover
            raise ValueError(
                "Agent runtime is not async compatible. Please use runtime_mode=async"
            )
        coro_or_future = self._agent.runtime.start_and_wait_completed()
        if iscoroutine(coro_or_future):
            return loop.create_task(coro_or_future)
        return cast(asyncio.Future, coro_or_future)


class AEARunner(AbstractMultipleRunner):
    """Run multiple AEA instances."""

    SUPPORTED_MODES: Dict[str, Type[AbstractMultipleExecutor]] = {
        "threaded": ThreadExecutor,
        "async": AsyncExecutor,
    }

    def __init__(
        self,
        agents: Sequence[AEA],
        mode: str,
        fail_policy: ExecutorExceptionPolicies = ExecutorExceptionPolicies.propagate,
    ) -> None:
        """
        Init AEARunner.

        :param agents: sequence of AEA instances to run.
        :param mode: executor name to use.
        :param fail_policy: one of ExecutorExceptionPolicies to be used with Executor
        """
        self._agents = agents
        super().__init__(mode=mode, fail_policy=fail_policy)

    def _make_tasks(self) -> Sequence[AbstractExecutorTask]:
        """Make tasks to run with executor."""
        return [AEAInstanceTask(agent) for agent in self._agents]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helper module 'dependency_tree'."""

import re
import shutil
import subprocess  # nosec
import tempfile
from collections import OrderedDict
from pathlib import Path

import pytest

from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.exceptions import AEAPackageLoadingError
from aea.helpers.base import cd
from aea.helpers.dependency_tree import DependencyTree, dump_yaml, load_yaml

from tests.conftest import PACKAGES_DIR


def test_generation_of_dependency_tree_of_repo_packages() -> None:
    """Test we can generate the dependency tree of the package directory of the repository."""
    dependency_tree = DependencyTree.generate(Path(PACKAGES_DIR))
    assert len(dependency_tree) > 0


def test_generation_of_dependency_tree_of_project() -> None:
    """Test we can generate the dependency tree of the project directory."""

    with tempfile.TemporaryDirectory() as tmp_dir:
        source_path = Path(PACKAGES_DIR)
        shutil.copytree(source_path, Path(tmp_dir) / source_path.parts[-1])
        with cd(tmp_dir):
            cmd = ["aea", "create", "dummy_project"]
            project_path = Path(tmp_dir) / "dummy_project"
            stdout, stderr = subprocess.Popen(cmd).communicate()  # nosec
            assert stdout is stderr is None
            dependency_tree = DependencyTree.generate(project_path, from_project=True)
            assert len(dependency_tree) > 0


def test_case_when_dependency_tree_has_a_self_loop() -> None:
    """Test we raise error when dependency tree has a node with a self-loop."""
    dummy_package_id = PackageId(
        PackageType.PROTOCOL, PublicId.from_str("author/pkg_name:0.1.0")
    )
    # self-loop: dummy_package -> dummy_package
    dependency_list = {dummy_package_id: [dummy_package_id]}
    with pytest.raises(
        AEAPackageLoadingError,
        match=re.escape(
            f"Found a self-loop dependency while resolving dependency tree in package {dummy_package_id}"
        ),
    ):
        DependencyTree.resolve_tree(dependency_list)


def test_case_when_dependency_tree_has_a_cycle() -> None:
    """Test we raise error when dependency tree has a cycle (length > 1)."""
    package_a_id = PackageId(
        PackageType.PROTOCOL, PublicId.from_str("author/pkg_a:0.1.0")
    )
    package_b_id = PackageId(
        PackageType.PROTOCOL, PublicId.from_str("author/pkg_b:0.1.0")
    )
    package_c_id = PackageId(
        PackageType.PROTOCOL, PublicId.from_str("author/pkg_c:0.1.0")
    )
    # cycle: package_a -> package_b -> package_c -> package_a
    dependency_list = {
        package_a_id: [package_b_id],
        package_b_id: [package_c_id],
        package_c_id: [package_a_id],
    }
    with pytest.raises(
        AEAPackageLoadingError,
        match=re.escape(
            f"Found a circular dependency while resolving dependency tree: {package_a_id} -> {package_b_id} -> {package_c_id} -> {package_a_id}"
        ),
    ):
        DependencyTree.resolve_tree(dependency_list)


@pytest.mark.parametrize("extra_data", [None, [{}] * 3])
def test_dump_yaml(extra_data) -> None:
    """Test dump yaml"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        filepath = Path(tmp_dir) / "dump_dummy.yaml"
        data = dict.fromkeys(reversed(range(3)))
        dump_yaml(filepath, data, extra_data)
        reconstituted_data, reconstituted_extra_data = load_yaml(filepath)
        assert isinstance(reconstituted_data, OrderedDict)
        assert list(reconstituted_data.keys()) == list(range(3))
        assert all(map(lambda o: isinstance(o, OrderedDict), reconstituted_extra_data))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Generate all the protocols from their specifications.

This script takes all the protocol specification (scraped from the protocol README)
and calls the `aea generate protocol` command.

Currently, it does a lot of assumptions, and might not be useful for
all use cases. However, with not much work, can be customized to achieve
the desired outcomes.

It requires the `aea` package, `black` and `isort` tools.
"""
import logging
import os
import pprint
import re
import shutil
import subprocess  # nosec
import sys
import tempfile
from distutils.dir_util import copy_tree  # pylint: disable=deprecated-module
from pathlib import Path
from typing import Any, List, Tuple, cast

import click
import semver

from aea.cli.registry.utils import download_file, extract, request_api
from aea.common import JSONLike
from aea.configurations.base import ComponentType, ProtocolConfig
from aea.configurations.constants import (
    AEA_TEST_DIRNAME,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_README_FILE,
    PROTOCOLS,
)
from aea.configurations.data_types import PackageId, PublicId
from aea.configurations.loader import ConfigLoaders, load_component_configuration
from aea.exceptions import enforce
from aea.helpers.git import check_working_tree_is_dirty
from aea.helpers.protocols import get_protocol_specification_from_readme
from aea.manager.helpers import AEAProject


SPECIFICATION_REGEX = re.compile(r"(---\nname.*\.\.\.)", re.DOTALL)
LIBPROTOC_VERSION = "libprotoc 24.3"
CUSTOM_TYPE_MODULE_NAME = "custom_types.py"
PROTOCOL_GENERATOR_DOCSTRING_REGEX = "It was created with protocol buffer compiler version `libprotoc .*` and aea protocol generator version `.*`."

logging.basicConfig(format="[%(asctime)s][%(levelname)s] %(message)s")
logger = logging.getLogger("generate_all_protocols")
logger.setLevel(logging.INFO)


def find_protocols_in_local_registry(packages_dir: Path) -> List[Path]:
    """Find all protocols in local registry."""
    return [
        p.parent.absolute()
        for p in packages_dir.glob(f"**/{DEFAULT_PROTOCOL_CONFIG_FILE}")
    ]


def log(message: str, level: int = logging.INFO) -> None:
    """Produce a logging message."""
    logger.log(level, message)


def log_separator(length: int = 100) -> None:
    """Log separator."""
    click.echo("=" * length)


def _save_specification_in_temporary_file(
    name: str, specification_content: str
) -> None:
    """
    Save the specification in a temporary file.

    :param name: the name of the package.
    :param specification_content: the specification content.
    """
    # here, the cwd is the temporary AEA project
    # hence, we are writing in a temporary directory
    spec_path = Path("..", name + ".yaml")
    log(f"Save specification '{name}' in temporary file {spec_path}")
    spec_path.write_text(specification_content)


def _generate_protocol(package_path: Path) -> None:
    """
    Generate the protocol.

    :param package_path: package to the path.
    """
    cmd = ["generate", "protocol", os.path.join("..", package_path.name) + ".yaml"]
    log(f"Generate the protocol. Command: {pprint.pformat(cmd)}")
    AEAProject.run_aea(*cmd)


def run_isort_and_black(directory: Path, **kwargs: Any) -> None:
    """Run black and isort against a directory."""
    kwargs["stdout"] = subprocess.PIPE

    try:
        # check if black and isort is installed
        import black  # type: ignore  # noqa  # pylint: disable=unused-import,import-outside-toplevel
        import isort  # type: ignore  # noqa  # pylint: disable=unused-import,import-outside-toplevel
    except ImportError as e:
        raise click.ClickException(str(e)) from e

    AEAProject.run_cli(
        sys.executable,
        "-m",
        "black",
        "-q",
        str(directory.absolute()),
        **kwargs,
    )
    AEAProject.run_cli(
        sys.executable,
        "-m",
        "isort",
        "--settings-path",
        "setup.cfg",
        str(directory.absolute()),
        **kwargs,
    )


def _fix_generated_protocol(package_path: Path) -> None:
    """
    Fix the generated protocol.

    That means:
    - replacing the prefix of import statements for default protocols;
    - restore the original custom types, if any.
    - copy the README, if any.

    :param package_path: path to the protocol package. Used also to recover the protocol name.
    """
    custom_types_module = package_path / CUSTOM_TYPE_MODULE_NAME
    if custom_types_module.exists():
        log(f"Restore original custom types in {package_path}")
        file_to_replace = Path(PROTOCOLS, package_path.name, CUSTOM_TYPE_MODULE_NAME)
        file_to_replace.write_text(custom_types_module.read_text())

    package_readme_file = package_path / DEFAULT_README_FILE
    if package_readme_file.exists():
        log(f"Copy the README {package_readme_file} into the new generated protocol.")
        shutil.copyfile(
            package_readme_file,
            Path(PROTOCOLS, package_path.name, DEFAULT_README_FILE),
        )

    tests_module = package_path / AEA_TEST_DIRNAME
    if tests_module.is_dir():
        log(f"Restore original `tests` directory in {package_path}")
        copy_tree(
            str(tests_module),
            str(Path(PROTOCOLS, package_path.name, AEA_TEST_DIRNAME)),
        )


def _update_original_protocol(package_path: Path) -> None:
    """
    Update the original protocol.

    :param package_path: the path to the original package. Used to recover the protocol name.
    """
    log(f"Copy the new protocol into the original directory {package_path}")
    shutil.rmtree(package_path)
    shutil.copytree(Path(PROTOCOLS, package_path.name), package_path)


def _fingerprint_protocol(name: str) -> None:
    """Fingerprint the generated (and modified) protocol."""
    log(f"Fingerprint the generated (and modified) protocol '{name}'")
    protocol_config = load_component_configuration(
        ComponentType.PROTOCOL,
        Path(PROTOCOLS, name),
        skip_consistency_check=True,
    )
    AEAProject.run_aea("fingerprint", "protocol", str(protocol_config.public_id))


def _parse_generator_docstring(package_path: Path) -> str:
    """
    Parse protocol generator docstring.

    The docstring this function searches is in the __init__.py module
    and it is of the form:

        It was created with protocol buffer compiler version `libprotoc ...` and aea version `...`.


    :param package_path: path to the protocol package
    :return: the docstring
    """
    content = (package_path / "__init__.py").read_text()
    regex = re.compile(PROTOCOL_GENERATOR_DOCSTRING_REGEX)
    match = regex.search(content)
    if match is None:
        raise ValueError("protocol generator docstring not found")
    return match.group(0)


def _replace_generator_docstring(package_path: Path, replacement: str) -> None:
    """
    Replace the generator docstring in the __init__.py module.

    (see _parse_generator_docstring for more details).

    :param package_path: path to the
    :param replacement: the replacement to use.
    """
    protocol_name = package_path.name
    init_module = Path(PROTOCOLS) / protocol_name / "__init__.py"
    content = init_module.read_text()
    content = re.sub(PROTOCOL_GENERATOR_DOCSTRING_REGEX, replacement, content)
    init_module.write_text(content)


def _check_preliminaries() -> None:
    """Check that the required packages are installed."""
    try:
        import aea  # noqa: F401  # pylint: disable=import-outside-toplevel,unused-import
    except ModuleNotFoundError:
        enforce(False, "'aea' package not installed.")
    enforce(shutil.which("black") is not None, "black command line tool not found.")
    enforce(shutil.which("isort") is not None, "isort command line tool not found.")
    enforce(shutil.which("protoc") is not None, "protoc command line tool not found.")
    result = subprocess.run(  # nosec
        ["protoc", "--version"], stdout=subprocess.PIPE, check=True
    )
    result_str = result.stdout.decode("utf-8")
    enforce(
        LIBPROTOC_VERSION in result_str,
        f"Invalid version for protoc. Found: {result_str}. Required: {LIBPROTOC_VERSION}.",
    )


def download_package(
    package_id: PackageId, destination_path: str
) -> None:  # pragma: nocover
    """Download a package into a directory."""
    api_path = f"/{package_id.package_type.to_plural()}/{package_id.author}/{package_id.name}/{package_id.public_id.LATEST_VERSION}"
    resp = cast(JSONLike, request_api("GET", api_path))
    file_url = cast(str, resp["file"])
    filepath = download_file(file_url, destination_path)
    extract(filepath, destination_path)


def _bump_protocol_specification_id(
    package_path: Path, configuration: ProtocolConfig
) -> None:
    """Bump spec id version."""
    spec_id: PublicId = configuration.protocol_specification_id  # type: ignore
    old_version = semver.VersionInfo.parse(spec_id.version)
    new_version = str(old_version.bump_minor())
    new_spec_id = PublicId(spec_id.author, spec_id.name, new_version)
    configuration.protocol_specification_id = new_spec_id
    with (package_path / DEFAULT_PROTOCOL_CONFIG_FILE).open("w") as file_output:
        ConfigLoaders.from_package_type(configuration.package_type).dump(
            configuration, file_output
        )


def _bump_protocol_specification_id_if_needed(package_path: Path) -> None:
    """
    Check if protocol specification id needs to be bumped.

    Workflow:
    - extract protocol specification file from README
    - download latest protocol id and extract its protocol specification as above
    - if different, bump protocol specification version, else don't.

    :param package_path: path to the protocol package.
    """
    # extract protocol specification file from README
    current_specification_content = get_protocol_specification_from_readme(package_path)

    # download latest protocol id and extract its protocol specification as above
    configuration: ProtocolConfig = cast(
        ProtocolConfig,
        load_component_configuration(ComponentType.PROTOCOL, package_path),
    )
    temp_directory = Path(tempfile.mkdtemp())
    try:
        download_package(configuration.package_id, str(temp_directory))
    except click.ClickException:
        log("Protocol specification id not bumped - new protocol.")
        return
    downloaded_package_directory = temp_directory / configuration.name
    old_specification_content = get_protocol_specification_from_readme(
        downloaded_package_directory
    )
    old_configuration: ProtocolConfig = cast(
        ProtocolConfig,
        load_component_configuration(
            ComponentType.PROTOCOL,
            downloaded_package_directory,
            skip_consistency_check=True,
        ),
    )

    # if different, bump protocol specification version, else don't.
    public_id_version_is_newer = (
        old_configuration.public_id.package_version  # type: ignore
        <= configuration.public_id.package_version
    )
    content_is_different = current_specification_content != old_specification_content
    if public_id_version_is_newer and content_is_different:
        log(
            f"Bumping protocol specification id from '{old_configuration.protocol_specification_id}' to '{configuration.protocol_specification_id}'"
        )
        _bump_protocol_specification_id(package_path, configuration)
        return
    log(
        "Protocol specification id not bumped - content is not different, or version is not newer."
    )


def replace_in_directory(name: str, replacement_pairs: List[Tuple[str, str]]) -> None:
    """
    Replace text in directory.

    :param name: the protocol name.
    :param replacement_pairs: a list of pairs of strings (to_replace, replacement).
    """
    log(f"Replace prefix of import statements in directory '{name}'")
    package_dir = Path(PROTOCOLS, name)
    for submodule in package_dir.rglob("*.py"):
        log(f"Process submodule {submodule.relative_to(package_dir)}")
        for to_replace, replacement in replacement_pairs:
            if to_replace not in submodule.read_text():
                continue
            submodule.write_text(submodule.read_text().replace(to_replace, replacement))


def _process_packages_protocol(
    package_path: Path,
    preserve_generator_docstring: bool,
    no_format: bool,
    root_dir: Path,
) -> None:
    """
    Process protocol package from local registry.

    If the flag '--no-bump' is specified, it means the protocol generator
    string that records the AEA and the protoc version used, i.e.:

        It was created with protocol buffer compiler version `libprotoc ...` and aea version `...`.

    must not be updated, as the 'AEA' version could have been changed.

    It means:
    - extract protocol specification from README
    - generate the protocol in the current AEA project
    - fix the generated protocol (e.g. import prefixed, custom types, ...)
    - update the original protocol with the newly generated one.

    It assumes the working directory is an AEA project.

    :param package_path: path to the package.
    :param preserve_generator_docstring: if True, the protocol generator docstring is preserved (see above).
    :param no_format: If True, won't run the formatters
    :param root_dir: Path to open-aea repository
    """
    if preserve_generator_docstring:
        # save the old protocol generator docstring
        old_protocol_generator_docstring = _parse_generator_docstring(package_path)
    specification_content = get_protocol_specification_from_readme(package_path)
    _save_specification_in_temporary_file(package_path.name, specification_content)
    _generate_protocol(package_path)
    _fix_generated_protocol(package_path)

    packages_dir_parent = package_path.parent.parent.parent.parent
    if root_dir != packages_dir_parent:
        if root_dir not in packages_dir_parent.parents:
            raise ValueError(
                "Packages dir should be a sub directory of the root directory."
            )

        relative_import_path = ".".join(
            packages_dir_parent.parts[len(root_dir.parts) :]
        )
        replacements = [
            (
                f"from packages.fetchai.protocols.{package_path.name}",
                f"from {relative_import_path}.packages.fetchai.protocols.{package_path.name}",
            )
        ]
        replace_in_directory(package_path.name, replacements)

    if preserve_generator_docstring:
        _replace_generator_docstring(package_path, old_protocol_generator_docstring)

    if not no_format:
        run_isort_and_black(Path(PROTOCOLS, package_path.name), cwd=str(root_dir))

    _fingerprint_protocol(package_path.name)
    _update_original_protocol(package_path)


@click.command(name="generate-all-protocols")
@click.argument(
    "packages_dir",
    type=click.Path(dir_okay=True, file_okay=False),
    default=Path("packages/"),
)
@click.option(
    "--no-bump",
    is_flag=True,
    default=False,
    help="no_bump implies to ignore the docstring",
)
@click.option(
    "--no-format",
    is_flag=True,
    default=False,
    help="If set to True, formatters won't run.",
)
@click.option(
    "--root-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    help="Path to root dir.",
    default=Path(".").resolve(),
)
@click.option(
    "--check-clean",
    is_flag=True,
    help="Check if Git tree is clean.",
)
def generate_all_protocols(
    packages_dir: Path,
    no_bump: bool,
    no_format: bool,
    root_dir: Path,
    check_clean: bool,
) -> None:
    """Generate all protocols."""

    _check_preliminaries()

    packages_dir = Path(packages_dir).absolute()
    root_dir = Path(root_dir).absolute()
    all_protocols = find_protocols_in_local_registry(packages_dir)

    with AEAProject():
        for package_path in all_protocols:
            log_separator()
            log(f"Processing protocol at path {package_path}")
            if not no_bump:
                _bump_protocol_specification_id_if_needed(package_path)
            _process_packages_protocol(
                package_path,
                no_bump,
                no_format,
                root_dir,
            )

    if check_clean:
        is_clean = check_working_tree_is_dirty()
        if not is_clean:
            print(
                "You should bump PROTOCOL_GENERATOR_VERSION in aea/protocols/__init__.py "
                "and rerun the protocol generation"
            )
            sys.exit(1)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains a custom crypto class for testing purposes."""
from typing import Any, Optional, Tuple, Union

from aea.common import Address
from aea.crypto.base import Crypto, EntityClass


class CustomCrypto(Crypto):
    """This is a custom crypto class for testing purposes.."""

    @classmethod
    def generate_private_key(  # type: ignore[empty-body]
        cls, extra_entropy: Union[str, bytes, int] = ""
    ) -> EntityClass:  # type: ignore[type-var]
        """Generare private key."""
        pass

    @classmethod
    def load_private_key_from_path(
        cls, file_name: str, password: Optional[str] = None
    ) -> Any:
        """
        Load a private key in hex format from a file.

        :param file_name: the path to the hex file.
        :param password: the password to encrypt/decrypt the private key.
        :return: the Entity.
        """
        pass

    @property
    def public_key(self) -> str:  # type: ignore[empty-body]
        """Get public key."""
        pass

    @property
    def address(self) -> str:  # type: ignore[empty-body]
        """Get address."""
        pass

    @property
    def private_key(self) -> str:  # type: ignore[empty-body]
        """Get private key."""
        pass

    @classmethod
    def get_address_from_public_key(cls, public_key: str) -> str:  # type: ignore[empty-body]
        """
        Get address from public key.

        :param public_key: the public key.
        :return: the address
        """
        pass

    def sign_message(self, message: bytes, is_deprecated_mode: bool = False) -> str:  # type: ignore[empty-body]
        """
        Sign message.

        :param message: the message
        :param is_deprecated_mode: whether or not deprecated signing mode is used.
        :return: signed message string
        """
        pass

    def sign_transaction(self, transaction: Any) -> Any:
        """
        Sign transaction.

        :param transaction: the transaction to be signed
        :return: the signed transaction
        """
        pass

    def recover_message(  # type: ignore[empty-body]
        self, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[Address, ...]:
        """
        Recover message.

        :param message: the message
        :param signature: the signature
        :param is_deprecated_mode: whether or not it is deprecated
        """
        pass

    def encrypt(self, password: str) -> str:  # type: ignore[empty-body]
        """
        Encrypt the private key and return in json.

        :param private_key: the raw private key.
        :param password: the password to decrypt.
        :return: json string containing encrypted private key.
        """

    @classmethod
    def decrypt(cls, keyfile_json: str, password: str) -> str:  # type: ignore[empty-body]
        """
        Decrypt the private key and return in raw form.

        :param keyfile_json: json string containing encrypted private key.
        :param password: the password to decrypt.
        :return: the raw private key.
        """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Package manager V1"""

import copy
import json
import logging
import traceback
from collections import OrderedDict
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Callable, Dict, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Union, cast

from requests import get as r_get

from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PackageId, PackageType
from aea.helpers.fingerprint import check_fingerprint
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.package_manager.base import (
    BasePackageManager,
    ConfigLoaderCallableType,
    PACKAGES_FILE,
    PACKAGE_SOURCE_RE,
    PackageFileNotValid,
    PackageIdToHashMapping,
    PackageNotValid,
    PackagesSourceNotValid,
    load_configuration,
)


GIT_TAGS_URL = "https://api.github.com/repos/{repo}/tags"
PACKAGE_FILE_REMOTE_URL = (
    "https://raw.githubusercontent.com/{repo}/{tag}/packages/packages.json"
)


class PackageManagerV1(BasePackageManager):
    """Package manager V1"""

    _third_party_packages: PackageIdToHashMapping
    _dev_packages: PackageIdToHashMapping

    class LocalPackageType(Enum):
        """Local package types."""

        DEV = "dev"
        THIRD_PARTY = "third_party"

    def __init__(
        self,
        path: Path,
        dev_packages: Optional[PackageIdToHashMapping] = None,
        third_party_packages: Optional[PackageIdToHashMapping] = None,
        config_loader: ConfigLoaderCallableType = load_configuration,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Initialize object."""
        super().__init__(
            path=path,
            config_loader=config_loader,
            logger=logger,
        )

        self._dev_packages = dev_packages or OrderedDict()
        self._third_party_packages = third_party_packages or OrderedDict()

    @property
    def dev_packages(
        self,
    ) -> PackageIdToHashMapping:
        """Returns mappings of package ids -> package hash"""

        return self._dev_packages

    @property
    def third_party_packages(
        self,
    ) -> PackageIdToHashMapping:
        """Returns mappings of package ids -> package hash"""

        return self._third_party_packages

    @property
    def all_packages(self) -> PackageIdToHashMapping:
        """Return all packages."""
        return OrderedDict(
            chain(self.dev_packages.items(), self.third_party_packages.items())
        )

    def get_package_hash(self, package_id: PackageId) -> Optional[str]:
        """Get package hash."""
        package_id = package_id.without_hash()
        return self.dev_packages.get(
            package_id, self.third_party_packages.get(package_id)
        )

    def is_third_party_package(self, package_id: PackageId) -> bool:
        """Check if a package is third party package."""

        return self._third_party_packages.get(package_id) is not None

    def is_dev_package(self, package_id: PackageId) -> bool:
        """Check if a package is third party package."""

        return self._dev_packages.get(package_id) is not None

    def add_package(
        self,
        package_id: PackageId,
        with_dependencies: bool = False,
        allow_update: bool = False,
    ) -> "PackageManagerV1":
        """Add package."""
        super().add_package(
            package_id=package_id,
            with_dependencies=with_dependencies,
            allow_update=allow_update,
        )
        self._third_party_packages[package_id] = self.calculate_hash_from_package_id(
            package_id
        )
        return self

    @staticmethod
    def _get_latest_tag(repo: str) -> str:
        """Get latest tag for the repository."""

        response = r_get(GIT_TAGS_URL.format(repo=repo))
        if response.status_code != 200:
            raise PackagesSourceNotValid(
                f"Fetching tags from `{repo}` failed with message '"
                + response.json()["message"]
                + "'"
            )

        latest_tag_data, *_ = response.json()
        return latest_tag_data["name"]

    @staticmethod
    def _get_packages_json(repo: str, tag: str) -> Dict[str, Dict[str, str]]:
        """Get `packages.json`."""

        response = r_get(PACKAGE_FILE_REMOTE_URL.format(repo=repo, tag=tag))
        if response.status_code != 200:
            raise PackagesSourceNotValid(
                f"Fetching packages from `{repo}` failed with message '"
                + response.text
                + "'"
            )
        return response.json()

    def _update_hashes_from_sources(self, sources: List[str]) -> None:
        """Update hashes from sources"""

        _updated_third_party_packages = copy.deepcopy(self.third_party_packages)

        self._logger.info("Fetching hashes from sources")
        for source in sources:
            source_regex = PACKAGE_SOURCE_RE.match(source)
            if source_regex is None:
                raise PackagesSourceNotValid(
                    f"Provided source name is not valid `{source}`"
                )

            repo, _, _, tag = source_regex.groups()
            if tag is None:
                tag = self._get_latest_tag(repo=repo)

            packages = self._get_packages_json(repo=repo, tag=tag)
            package_manager = self.from_json(packages=packages)

            for package_id in _updated_third_party_packages:
                latest_hash = package_manager.dev_packages.get(package_id)
                if latest_hash is None:
                    continue

                self._logger.info(
                    f"Found hash for third party package {package_id} in source `{repo}`, updating hash"
                )
                _updated_third_party_packages[package_id] = latest_hash

        self._third_party_packages = _updated_third_party_packages
        self.dump(file=self._packages_file)

    def register(
        self,
        package_path: Path,
        package_type: Optional[PackageType] = None,
    ) -> "PackageManagerV1":
        """Add package to the index."""
        package_type = package_type or PackageType(package_path.parent.name[:-1])
        package_config = self.config_loader(package_type, package_path)
        self.update_fingerprints(package_id=package_config.package_id)
        self._dev_packages[package_config.package_id] = IPFSHashOnly.hash_directory(
            dir_path=str(package_path)
        )
        return self

    def sync(
        self,
        dev: bool = False,
        third_party: bool = True,
        update_packages: bool = False,
        update_hashes: bool = False,
        sources: Optional[List[str]] = None,
    ) -> "PackageManagerV1":
        """Sync local packages to the remote registry."""

        if update_packages and update_hashes:
            raise ValueError(
                "Both `update_packages` and `update_hashes` cannot be set to `True`."
            )
        self._logger.info(f"Performing sync @ {self.path}")

        if sources is not None and len(sources) > 0:
            self._update_hashes_from_sources(sources=sources)

        sync_needed = False
        if third_party:
            self._logger.info("Checking third party packages.")
            (
                _sync_needed,
                hash_updates_third_party,
                package_updates_third_party,
            ) = self._sync(
                packages=self.third_party_packages,
                update_hashes=update_hashes,
                update_packages=update_packages,
            )

            sync_needed = sync_needed or _sync_needed
            if update_hashes and hash_updates_third_party:
                third_party_package_id = "\n\t- ".join(
                    map(str, hash_updates_third_party)
                )
                self._logger.warning(
                    f"Hashes for follwing third party module has changed.\n\t- {third_party_package_id}"
                )

            if update_packages and package_updates_third_party:
                self._logger.info("Updating third party packages.")
                for package_id, _ in package_updates_third_party.items():
                    self._logger.info(f"Updating {package_id}")
                    self.update_package(package_id=package_id)

        if dev:
            self._logger.info("Checking dev packages.")
            _sync_needed, hash_updates_dev, package_updates_dev = self._sync(
                packages=self.dev_packages,
                update_hashes=update_hashes,
                update_packages=update_packages,
            )
            sync_needed = sync_needed or _sync_needed

            if update_hashes:
                self._dev_packages = hash_updates_dev

            if update_packages and package_updates_dev:
                self._logger.info("Updating dev packages.")
                for package_id, _ in package_updates_dev.items():
                    self._logger.info(f"Updating {package_id}")
                    self.update_package(package_id=package_id)

        if update_hashes:
            self._logger.info("Updating hashes")
            self.dump()

        if sync_needed:
            self._logger.info("Sync complete")
            self._logger.info("Updating hashes")
            self.dump()
        else:
            self._logger.info("No package was updated.")

        return self

    def update_package_hashes(
        self,
        selector_prompt: Optional[Callable[[], str]] = None,
        skip_missing: bool = False,
    ) -> "PackageManagerV1":
        """Update package.json file."""

        for package_id in self.iter_dependency_tree():
            self.update_fingerprints(package_id=package_id)
            self.update_dependencies(package_id=package_id)

            package_hash = self.calculate_hash_from_package_id(package_id=package_id)
            is_dev_package = self.is_dev_package(package_id=package_id)
            is_third_party_package = self.is_third_party_package(package_id=package_id)

            if is_dev_package:
                if self._dev_packages[package_id] == package_hash:
                    continue

                self._logger.info(f"Updating hash for {package_id}")
                self._dev_packages[package_id] = package_hash
                continue

            if is_third_party_package:
                if self._third_party_packages[package_id] == package_hash:
                    continue

                self._third_party_packages[package_id] = package_hash
                self._logger.warning(
                    "Hash change detected for third party package"
                    f"\n\tPackage: {package_id}"
                    f"\n\tCalculated hash: {package_hash}"
                    f"\n\tExpected hash: {self._third_party_packages[package_id]}"
                )
                continue

            if skip_missing:
                continue

            if selector_prompt is None:
                raise PackageNotValid(
                    f"Found a package which is not listed in the `packages.json` with package id {package_id}"
                )

            self._logger.info(f"A new package found with package ID {package_id}")
            package_type = self.LocalPackageType(selector_prompt())
            if package_type == self.LocalPackageType.DEV:
                self._logger.info("Adding package to dev packages")
                self._dev_packages[package_id] = package_hash
            else:
                self._logger.info("Adding package to third party packages")
                self._third_party_packages[package_id] = package_hash

        return self

    @staticmethod
    def _calculate_hash(package_path: Union[Path, str]) -> str:
        """Calculate hash for path."""
        return IPFSHashOnly.get(str(package_path))

    def verify(
        self,
    ) -> int:
        """Verify fingerprints and outer hash of all available packages."""

        failed = False
        try:
            for package_id in self.iter_dependency_tree():
                self._logger.info(f"Verifying {package_id}")
                package_path = self.package_path_from_package_id(package_id)
                configuration_obj = self._get_package_configuration(
                    package_id=package_id
                )

                calculated_hash = self._calculate_hash(str(package_path))
                fingerprint_check = check_fingerprint(configuration_obj)

                if not fingerprint_check:
                    failed = True
                    self._logger.error(
                        f"Fingerprints does not match for {package_id} @ {package_path}"
                    )
                    if self.is_third_party_package(package_id=package_id):
                        self._third_party_packages[package_id] = calculated_hash
                    continue

                expected_hash = self.dev_packages.get(
                    package_id, self.third_party_packages.get(package_id)
                )
                if expected_hash is None:
                    failed = True
                    self._logger.error(f"Cannot find hash for {package_id}")
                    continue
                if calculated_hash != expected_hash:
                    # this update is required for further dependency checks for
                    # packages where this package is a dependency
                    if self.is_third_party_package(package_id=package_id):
                        self._third_party_packages[package_id] = calculated_hash
                    if self.is_dev_package(package_id=package_id):
                        self._dev_packages[package_id] = calculated_hash
                    self._logger.error(f"Hash does not match for {package_id}")
                    self._logger.error(f"\tCalculated hash: {calculated_hash}")
                    self._logger.error(f"\tExpected hash: {expected_hash}")
                    failed = True
                    continue

                if not self.is_dependencies_hashes_match(package_id, configuration_obj):
                    failed = True

        except Exception:  # pylint: disable=broad-except
            traceback.print_exc()
            failed = True

        return int(failed)

    @property
    def json(
        self,
    ) -> OrderedDictType:
        """Json representation"""

        data: OrderedDictType[str, OrderedDictType[str, str]] = OrderedDict()
        data["dev"] = OrderedDict()
        data["third_party"] = OrderedDict()

        for package_id, package_hash in self._dev_packages.items():
            data["dev"][package_id.to_uri_path] = package_hash

        for package_id, package_hash in self._third_party_packages.items():
            data["third_party"][package_id.to_uri_path] = package_hash

        return data

    @staticmethod
    def _load_packages(packages_file: Path) -> Dict[str, Dict[str, str]]:
        """Load packages json file."""
        return cast(Dict, json.loads(packages_file.read_text()))

    @classmethod
    def from_json(
        cls,
        packages: Dict[str, Dict[str, str]],
        packages_dir: Optional[Path] = None,
        config_loader: ConfigLoaderCallableType = load_configuration,
    ) -> "PackageManagerV1":
        """Initialize from json object"""

        if "dev" not in packages:
            raise PackageFileNotValid(
                "`packages.json` file not valid, no development packages found"
            )

        if "third_party" not in packages:
            raise PackageFileNotValid(
                "`packages.json` file not valid, no third party packages found"
            )

        dev_packages = OrderedDict()
        for package_id, package_hash in packages["dev"].items():
            dev_packages[PackageId.from_uri_path(package_id)] = package_hash

        third_party_packages = OrderedDict()
        for package_id, package_hash in packages["third_party"].items():
            third_party_packages[PackageId.from_uri_path(package_id)] = package_hash

        return cls(
            path=(packages_dir or Path.cwd() / PACKAGES),
            dev_packages=dev_packages,
            third_party_packages=third_party_packages,
            config_loader=config_loader,
        )

    @classmethod
    def from_dir(
        cls,
        packages_dir: Path,
        config_loader: ConfigLoaderCallableType = load_configuration,
    ) -> "PackageManagerV1":
        """Initialize from packages directory."""
        packages_file = packages_dir / PACKAGES_FILE
        packages = cls._load_packages(packages_file)
        return cls.from_json(
            packages=packages,
            packages_dir=packages_dir,
            config_loader=config_loader,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains tests for aea.cli.add generic methods."""

import os
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase, mock

import click
import pytest

from aea.cli.add import _add_item_deps
from aea.cli.core import cli
from aea.cli.registry.settings import REMOTE_IPFS
from aea.configurations.data_types import PackageType, PublicId
from aea.test_tools.click_testing import CliRunner
from aea.test_tools.test_cases import AEATestCaseEmpty, CLI_LOG_OPTION

from tests.conftest import AUTHOR, MAX_FLAKY_RERUNS, ROOT_DIR, TEST_IPFS_REGISTRY_CONFIG
from tests.test_cli.tools_for_testing import ContextMock


class AddItemDepsTestCase(TestCase):
    """Test case for _add_item_deps method."""

    @mock.patch("aea.cli.add.add_item")
    def test__add_item_deps_missing_skills_positive(self, add_item_mock):
        """Test _add_item_deps for positive result with missing skills."""
        ctx = ContextMock(skills=[])
        item_config = mock.Mock()
        item_config.protocols = []
        item_config.contracts = []
        item_config.connections = []
        item_config.skills = ["skill-1", "skill-2"]
        _add_item_deps(ctx, "skill", item_config)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
class BaseTestAddRemoteMode(AEATestCaseEmpty):
    """Base test case for add component, --remote mode."""

    IS_EMPTY = True
    COMPONENT_ID = PublicId(
        "valory",
        "test_abci",
        "0.1.0",
        "bafybeigb5myrmvjbfcdcwodoxn7ijdinntza3fagwwjbiho7kjdcr5lbmm",
    )
    COMPONENT_TYPE = PackageType.SKILL

    @mock.patch(
        "aea.cli.registry.utils.get_or_create_cli_config",
        return_value=TEST_IPFS_REGISTRY_CONFIG,
    )
    @mock.patch("aea.cli.add.get_default_remote_registry", return_value=REMOTE_IPFS)
    def test_add_component_remote_mode(self, *_):
        """Test add skill mixed mode."""
        self.run_cli_command(
            "add",
            "--remote",
            str(self.COMPONENT_TYPE),
            str(self.COMPONENT_ID),
            cwd=self._get_cwd(),
        )

        items_path = os.path.join(
            self.agent_name,
            "vendor",
            self.COMPONENT_ID.author,
            self.COMPONENT_TYPE.to_plural(),
        )
        items_folders = os.listdir(items_path)
        item_name = self.COMPONENT_ID.name
        assert item_name in items_folders


class BaseTestAddConnectionLocalWhenNoLocalRegistryExists:
    """Test that the command 'aea add connection' fails in local mode when the local registry does not exists."""

    COMPONENT_ID: PublicId
    COMPONENT_TYPE: PackageType

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )
        assert result.exit_code == 0, result.stdout

        result = cls.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "--registry-path",
                str(Path(ROOT_DIR) / "packages"),
                "create",
                cls.agent_name,
                "--local",
            ],
            standalone_mode=False,
        )
        assert result.exit_code == 0, result.stdout

        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "add",
                "--local",
                str(cls.COMPONENT_TYPE),
                str(cls.COMPONENT_ID),
            ],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1."""
        assert self.result.exit_code == 1, self.result.stdout

    def test_standard_output_mentions_failure(self):
        """Test standard output contains information on failure."""
        assert (
            "Registry path not provided and local registry `packages` not found in current (.) and parent directory."
            in self.result.exception.message
        )

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class BaseTestAddSkillMixedModeFallsBack(AEATestCaseEmpty):
    """Test add skill in mixed mode that fails with local falls back to remote registry."""

    IS_EMPTY = True
    COMPONENT_ID: PublicId
    COMPONENT_TYPE: PackageType

    @mock.patch(
        "aea.cli.add.find_item_locally_or_distributed",
        side_effect=click.ClickException(""),
    )
    @mock.patch(
        "aea.cli.add.load_item_config",
    )
    @mock.patch(
        "aea.cli.add.register_item",
    )
    @mock.patch(
        "aea.cli.add.is_fingerprint_correct",
    )
    @mock.patch(
        "aea.cli.add.PublicId.from_json",
    )
    def test_add_skill_remote_mode_negative_local_positive_remote(self, *_mocks):
        """Test add skill mixed mode."""
        with mock.patch(
            "aea.cli.add.fetch_item_remote",
        ) as mock_fetch_item_remote:
            self.run_cli_command(
                "add",
                str(self.COMPONENT_TYPE),
                "--mixed",
                str(self.COMPONENT_ID),
                cwd=self._get_cwd(),
            )

        mock_fetch_item_remote.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Ethereum module wrapping the public and private key cryptography and ledger api."""

import logging
import warnings
from typing import Any, Optional, Union, cast

from aea_ledger_ethereum.ethereum import (
    AttributeDictTranslator as BaseAttributeDictTranslator,
)
from aea_ledger_ethereum.ethereum import EthereumApi, EthereumFaucetApi, EthereumHelper
from aea_ledger_ethereum.ethereum import (
    SignedTransactionTranslator as BaseSignedTransactionTranslator,
)
from aea_ledger_ethereum.ethereum import (
    TESTNET_NAME,
    set_wrapper_for_web3py_session_cache,
)
from aea_ledger_ethereum_hwi.account import HWIAccount
from eth_account.messages import encode_defunct
from eth_account.signers.local import LocalAccount

from aea.common import JSONLike
from aea.crypto.base import Crypto


_default_logger = logging.getLogger("aea.ledger_apis.ethereum_hwi")

SignedTransactionTranslator = BaseSignedTransactionTranslator
AttributeDictTranslator = BaseAttributeDictTranslator

_ETHEREUM_HWI = "ethereum_hwi"
DEFAULT_DEVICE_INDEX = 0
DEFAULT_KEYPAIR_INDEX = 0
DEFAULT_ACCOUNT_INDEX = 0


class EthereumHWICrypto(Crypto[HWIAccount]):
    """Class wrapping the Account Generation from Ethereum ledger."""

    identifier = _ETHEREUM_HWI

    def __init__(
        self,
        private_key_path: Optional[str] = None,
        password: Optional[str] = None,
        extra_entropy: Union[str, bytes, int] = "",
        **kwargs: Any,
    ) -> None:
        """
        Instantiate an ethereum crypto object.

        :param private_key_path: the private key path of the agent
        :param password: the password to encrypt/decrypt the private key.
        :param extra_entropy: add extra randomness to whatever randomness your OS can provide
        :param kwargs: extra keyword arguments
        """
        self._device_index = kwargs.pop(
            "default_device_index",
            DEFAULT_DEVICE_INDEX,
        )
        self._account_index = kwargs.pop(
            "default_account_index",
            DEFAULT_ACCOUNT_INDEX,
        )
        self._keypair_index = kwargs.pop(
            "default_keypair_index",
            DEFAULT_KEYPAIR_INDEX,
        )
        account = HWIAccount(
            default_device_index=self._device_index,
            default_account_index=self._account_index,
            default_key_index=self._keypair_index,
        )
        super().__init__(
            private_key_path=private_key_path,
            password=password,
            extra_entropy=extra_entropy,
            entity=account,
        )

    @property
    def private_key(self) -> str:
        """
        Return a private key.

        64 random hex characters (i.e. 32 bytes) + "0x" prefix.
        """

        raise NotImplementedError()

    @property
    def public_key(self) -> str:
        """
        Return a public key in hex format.

        128 hex characters (i.e. 64 bytes) + "0x" prefix.

        :return: a public key string in hex format
        """
        return self.entity.public_key

    @property
    def address(self) -> str:
        """
        Return the address for the key pair.

        40 hex characters (i.e. 20 bytes) + "0x" prefix.

        :return: an address string in hex format
        """
        return self.entity.address

    @classmethod
    def load_private_key_from_path(
        cls, file_name: str, password: Optional[str] = None
    ) -> LocalAccount:
        """
        Load a private key in hex format from a file.

        :param file_name: the path to the hex file.
        :param password: the password to encrypt/decrypt the private key.
        """

        raise NotImplementedError()

    def sign_message(self, message: bytes, is_deprecated_mode: bool = False) -> str:
        """
        Sign a message in bytes string form.

        :param message: the message to be signed
        :param is_deprecated_mode: if the deprecated signing is used
        :return: signature of the message in string form
        """
        if is_deprecated_mode and len(message) == 32:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                signature_dict = cast(HWIAccount, self.entity).signHash(message)
            signed_msg = signature_dict["signature"].hex()
        else:
            signable_message = encode_defunct(primitive=message)
            signature = cast(HWIAccount, self.entity).sign_message(
                signable_message=signable_message
            )
            signed_msg = signature["signature"].hex()
        return signed_msg

    def sign_transaction(self, transaction: JSONLike, **kwargs: Any) -> JSONLike:
        """
        Sign a transaction in bytes string form.

        :param transaction: the transaction to be signed
        :param kwargs: extra keyword arguments
        :return: signed transaction
        """
        signed_transaction = cast(HWIAccount, self.entity).sign_transaction(
            transaction_dict=transaction, **kwargs
        )
        signed_transaction_dict = SignedTransactionTranslator.to_dict(
            signed_transaction
        )
        return cast(JSONLike, signed_transaction_dict)

    @classmethod
    def generate_private_key(
        cls, extra_entropy: Union[str, bytes, int] = ""
    ) -> LocalAccount:
        """
        Generate a key pair for ethereum network.

        :param extra_entropy: add extra randomness to whatever randomness your OS can provide
        """

        raise NotImplementedError()

    def encrypt(self, password: str) -> str:
        """
        Encrypt the private key and return in json.

        :param password: the password to decrypt.
        """

        raise NotImplementedError()

    @classmethod
    def decrypt(cls, keyfile_json: str, password: str) -> str:
        """
        Decrypt the private key and return in raw form.

        :param keyfile_json: json str containing encrypted private key.
        :param password: the password to decrypt.
        """

        raise NotImplementedError()


class EthereumHWIHelper(EthereumHelper):
    """Helper class usable as Mixin for EthereumApi or as standalone class."""


class EthereumHWIApi(EthereumApi, EthereumHWIHelper):
    """Class to interact with the Ethereum Web3 APIs."""

    identifier = _ETHEREUM_HWI

    def __init__(self, **kwargs: Any):
        """Initialize object."""
        super().__init__(**kwargs)


class EthereumHWIFaucetApi(EthereumFaucetApi):
    """Ethereum testnet faucet API."""

    identifier = _ETHEREUM_HWI
    testnet_name = TESTNET_NAME


set_wrapper_for_web3py_session_cache()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for CLI test command."""
import shutil
import subprocess  # nosec
import sys
from pathlib import Path
from textwrap import dedent
from typing import Any, Sequence, Type
from unittest import mock

import click.testing
import pytest
from _pytest.config import ExitCode  # type: ignore

from aea.cli import cli
from aea.cli.test import get_packages_list
from aea.cli.utils.package_utils import get_package_path
from aea.configurations.constants import AEA_TEST_DIRNAME
from aea.configurations.data_types import (
    ComponentType,
    PackageId,
    PackageType,
    PublicId,
)
from aea.helpers.base import cd
from aea.package_manager.v1 import PackageManagerV1
from aea.test_tools.test_cases import AEATestCaseEmpty, CLI_LOG_OPTION


OK_PYTEST_EXIT_CODE = ExitCode.OK
NO_TESTS_COLLECTED_PYTEST_EXIT_CODE = ExitCode.NO_TESTS_COLLECTED


def _parametrize_class(test_cls: Type) -> Type:
    """Allow a base AEA test class to use 'setup/teardown' instead of '(setup/teardown)_class' methods."""
    old_setup_class = test_cls.setup_class
    old_teardown_class = test_cls.teardown_class

    def setup_class(cls) -> None:
        """Don't call super setup method."""

    def teardown_class(cls) -> None:
        """Don't call super teardown method."""

    def setup(self) -> None:
        """Setup after every test execution."""
        old_setup_class()  # type: ignore

    def teardown(self) -> None:
        """Tear down after every test execution."""
        old_teardown_class()  # type: ignore

    test_cls.setup_class = setup_class
    test_cls.teardown_class = teardown_class
    test_cls.setup = setup
    test_cls.teardown = teardown

    return test_cls


def mock_pytest_main() -> Any:
    """
    Mock pytest main so to run in a subprocess.

    This is necessary as otherwise the in-process subcall to pytest will conflict with the current pytest runtime.
    """

    def fun(args) -> int:
        result = subprocess.call(["pytest", *args])  # nosec
        return result

    return mock.patch("aea.cli.test.pytest.main", side_effect=fun)


class BaseAEATestCommand(AEATestCaseEmpty):
    """Base class for tests related to the command `aea test`."""

    @classmethod
    def run_test_command(cls, *args: Sequence[str]) -> click.testing.Result:
        """
        Execute an 'aea test' command.

        It differs from the method run_cli_command of the base class as it does not
        raise exception when the executed CLI command fails.

        :param args: the argumnets to pass to `aea test`
        :return: the click.testing.Result object
        """
        with cd(cls._get_cwd()):
            return cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "test", *args],
                standalone_mode=False,
                catch_exceptions=False,
            )

    def get_aea_dirpath(self) -> Path:
        """Get the AEA directory path."""
        return self.t / self.agent_name

    def get_test_aea_dirpath(self) -> Path:
        """Get the test AEA directory path."""
        return self.get_aea_dirpath() / AEA_TEST_DIRNAME

    def dummy_package_dirpath(
        self, package_type: ComponentType, item_name: str
    ) -> Path:
        """
        Get the package directory path.

        :param package_type: the type of the package
        :param item_name: the name of the item
        :return: path to the AEA package
        """
        public_id = self._public_id(package_type)
        result = get_package_path(
            str(self.get_aea_dirpath()),
            str(package_type.value),
            public_id,
            is_vendor=False,
        )
        return Path(result)

    def get_test_package_dirpath(
        self, package_type: ComponentType, item_name: str
    ) -> Path:
        """
        Get the test package directory path.

        :param package_type: the type of the package
        :param item_name: the name of the item
        :return: path to the AEA package
        """
        return self.dummy_package_dirpath(package_type, item_name) / AEA_TEST_DIRNAME

    @classmethod
    def write_dummy_test_module(cls, path_to_module: Path) -> None:
        """Write dummy test module."""
        path_to_module.write_text(
            dedent(
                """\
        def test_dummy_function():
            assert True
        """
            )
        )

    @classmethod
    def _get_dummy_package_name(cls, package_type: ComponentType) -> str:
        """Get the name of the dummy package from package type."""
        return f"dummy_{package_type.value}"

    def _scaffold_item(self, package_type: ComponentType) -> None:
        """Scaffold an item for testing."""
        item_name = self._get_dummy_package_name(package_type)
        self.scaffold_item(str(package_type.value), item_name)
        package_dirpath = self.dummy_package_dirpath(package_type, item_name)
        # initialize tests folder
        (package_dirpath / AEA_TEST_DIRNAME).mkdir(exist_ok=False)

    def _public_id(self, package_type: ComponentType) -> PublicId:
        """Return the PublicId of the dummy package."""
        return PublicId(self.author, self._get_dummy_package_name(package_type))

    def _configure_package_for_testing(self, package_type: ComponentType) -> None:
        """Configure a package for testing."""
        self._scaffold_item(package_type)
        test_package_name = self._get_dummy_package_name(package_type)
        test_package_dirpath = self.get_test_package_dirpath(
            package_type, test_package_name
        )
        test_module_filepath = (
            test_package_dirpath / f"test_{package_type.value}_module.py"
        )
        self.write_dummy_test_module(test_module_filepath)
        self.fingerprint_item(
            str(package_type.value), str(self._public_id(package_type))
        )

    def _vendorize(self, package_type: ComponentType, public_id: PublicId) -> None:
        """Vendorize an agent package."""
        aea_dirpath = self.get_aea_dirpath()
        old_package_dirpath = Path(
            get_package_path(
                str(aea_dirpath),
                str(package_type.value),
                public_id,
                is_vendor=False,
            )
        )
        new_package_dirpath = Path(
            get_package_path(
                str(aea_dirpath),
                str(package_type.value),
                public_id,
                is_vendor=True,
            )
        )
        assert old_package_dirpath.exists()
        assert not new_package_dirpath.exists()
        new_package_dirpath.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(str(old_package_dirpath), str(new_package_dirpath))


class TestAgentTestEmptySuite(BaseAEATestCommand):
    """Test that the command 'aea test' works as expected (with an empty test suite)."""

    def test_run(self, mock_sys_modules):
        """Assert that the exit code is equal to 5 (i.e. pytest succeeds without collecting tests)."""
        (self.get_aea_dirpath() / AEA_TEST_DIRNAME).mkdir(exist_ok=False)
        result = self.run_test_command()
        assert result.exit_code == NO_TESTS_COLLECTED_PYTEST_EXIT_CODE


class TestAgentTestSingleTest(BaseAEATestCommand):
    """Test that the command 'aea test' works as expected (with a non-empty test suite)."""

    def test_run(self, mock_sys_modules) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        # write dummy test module in test/ folder
        self.get_test_aea_dirpath().mkdir(exist_ok=False)
        test_module_filepath = self.get_test_aea_dirpath() / "test_module.py"
        self.write_dummy_test_module(test_module_filepath)
        result = self.run_test_command()
        assert result.exit_code == OK_PYTEST_EXIT_CODE


class TestAgentTestByPathSingleTest(BaseAEATestCommand):
    """Test that the command 'aea test by-path' works as expected (with a non-empty test suite)."""

    def test_run(self, mock_sys_modules) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        # write dummy test module in test/ folder
        self.get_test_aea_dirpath().mkdir(exist_ok=False)
        test_module_filepath = self.get_test_aea_dirpath() / "test_module.py"
        self.write_dummy_test_module(test_module_filepath)
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "test", "by-path", self._get_cwd()],
            standalone_mode=False,
            catch_exceptions=False,
        )

        assert result.exit_code == OK_PYTEST_EXIT_CODE


@_parametrize_class
class TestPackageTestByTypeEmptyTestSuite(BaseAEATestCommand):
    """Test that the command 'aea test item_type public_id' works as expected (with an empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(self, package_type: ComponentType, mock_sys_modules) -> None:
        """Assert that the exit code is equal to 5 (empty test suite)."""
        self._scaffold_item(package_type)
        public_id = self._public_id(package_type)
        result = self.run_test_command(str(package_type.value), str(public_id))
        assert result.exit_code == NO_TESTS_COLLECTED_PYTEST_EXIT_CODE


@_parametrize_class
class TestPackageTestByType(BaseAEATestCommand):
    """Test that the command 'aea test item_type public_id' works as expected (with a non-empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(
        self, package_type: ComponentType, mock_sys_modules, *_mocks: Any
    ) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        self._configure_package_for_testing(package_type)
        public_id = self._public_id(package_type)
        with mock_pytest_main():
            result = self.run_test_command(str(package_type.value), str(public_id))
            assert result.exit_code == OK_PYTEST_EXIT_CODE


@_parametrize_class
class TestVendorPackageTestByTypeEmptyTestSuite(BaseAEATestCommand):
    """Test that the command 'aea test item_type public_id' for vendor packages works as expected (with an empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(self, package_type: ComponentType, mock_sys_modules) -> None:
        """Assert that the exit code is equal to 5 (empty test suite)."""
        self._scaffold_item(package_type)
        public_id = self._public_id(package_type)
        self._vendorize(package_type, public_id)
        result = self.run_test_command(str(package_type.value), str(public_id))
        assert result.exit_code == NO_TESTS_COLLECTED_PYTEST_EXIT_CODE


@_parametrize_class
class TestVendorPackageTestByType(BaseAEATestCommand):
    """Test that the command 'aea test item_type public_id' for vendor packages works as expected (with a non-empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(
        self, package_type: ComponentType, mock_sys_modules, *_mocks: Any
    ) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        self._configure_package_for_testing(package_type)
        public_id = self._public_id(package_type)
        self._vendorize(package_type, public_id)
        with mock_pytest_main():
            result = self.run_test_command(str(package_type.value), str(public_id))
            assert result.exit_code == OK_PYTEST_EXIT_CODE
            # assert the module packages have been loaded
            assert (
                f"packages.default_author.{package_type.to_plural()}.dummy_{package_type.value}"
                in sys.modules
            )


@_parametrize_class
class TestPackageTestByPathEmptyTestSuite(BaseAEATestCommand):
    """Test that the command 'aea test by-path path-to-package' works as expected (empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(self, package_type: ComponentType, mock_sys_modules) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        self._scaffold_item(package_type)
        test_package_name = self._get_dummy_package_name(package_type)
        package_dirpath = self.dummy_package_dirpath(package_type, test_package_name)
        with mock_pytest_main():
            result = self.run_test_command(
                "by-path",
                str(package_dirpath),
            )
            assert result.exit_code == NO_TESTS_COLLECTED_PYTEST_EXIT_CODE


@_parametrize_class
class TestPackageTestByPath(BaseAEATestCommand):
    """Test that the command 'aea test by-path path-to-package' works as expected (non-empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(
        self, package_type: ComponentType, mock_sys_modules, *_mocks: Any
    ) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        self._configure_package_for_testing(package_type)
        test_package_name = self._get_dummy_package_name(package_type)
        package_dirpath = self.dummy_package_dirpath(package_type, test_package_name)
        with mock_pytest_main():
            result = self.run_test_command(
                "by-path",
                str(package_dirpath),
            )
            assert result.exit_code == OK_PYTEST_EXIT_CODE


@_parametrize_class
class TestPackageTestByPathWithCov(BaseAEATestCommand):
    """Test that the command 'aea test by-path path-to-package' works as expected (non-empty test suite)."""

    @pytest.mark.parametrize("package_type", list(ComponentType))
    def test_run(
        self, package_type: ComponentType, mock_sys_modules, *_mocks: Any
    ) -> None:
        """Assert that the exit code is equal to 0 (tests are run successfully)."""
        self._configure_package_for_testing(package_type)
        test_package_name = self._get_dummy_package_name(package_type)
        package_dirpath = self.dummy_package_dirpath(package_type, test_package_name)
        with mock_pytest_main():
            result = self.run_test_command(
                "--cov",
                "by-path",
                str(package_dirpath),
            )
            assert result.exit_code == OK_PYTEST_EXIT_CODE

            agent_path = self.t / self.agent_name

            assert (agent_path / "coverage.xml").exists()
            assert (agent_path / "htmlcov").exists()
            assert (agent_path / ".coverage").exists()


def test_get_packages_list() -> None:
    """Test get packages list function."""
    package_manager = PackageManagerV1(Path("some"))
    dev_package = PackageId(
        package_type=PackageType.CONNECTION,
        public_id=PublicId(author="test", name="third", version="1.0.0"),
    )
    third_part_package = PackageId(
        package_type=PackageType.SKILL,
        public_id=PublicId(author="test", name="third", version="1.0.0"),
    )
    package_manager._dev_packages = {dev_package: "some_hash"}  # type: ignore
    package_manager._third_party_packages = {third_part_package: "some_hash"}  # type: ignore

    with mock.patch.object(
        PackageManagerV1, "from_dir", return_value=package_manager
    ), mock.patch.object(
        package_manager, "package_path_from_package_id", return_value=Path("some")
    ):
        assert len(get_packages_list(Path("some"), packages_filter="dev")) == 1
        assert (
            get_packages_list(Path("some"), packages_filter="dev")[0][0]
            == PackageType.CONNECTION.to_plural()
        )
        assert len(get_packages_list(Path("some"), packages_filter="all")) == 2

        with pytest.raises(
            ValueError, match="Unknown package filter: bad. Valid are all,dev"
        ):
            get_packages_list(Path("some"), packages_filter="bad")


class TestPackageTestPackages(BaseAEATestCommand):
    """Test that the command 'aea test pacakages' works as expected (non-empty test suite)."""

    def test_packages_all_dev(self) -> None:
        """Check test pacakges --all  flag on/off."""

        with mock_pytest_main(), mock.patch(
            "aea.cli.test.get_packages_list", return_value=[]
        ) as get_packages_list_mock:
            result = self.run_test_command(
                "packages",
            )
            assert result.exit_code == OK_PYTEST_EXIT_CODE
            get_packages_list_mock.assert_called_once_with(
                packages_dir=mock.ANY, packages_filter="dev"
            )

        with mock_pytest_main(), mock.patch(
            "aea.cli.test.get_packages_list", return_value=[]
        ) as get_packages_list_mock:
            result = self.run_test_command("packages", "--all")
            assert result.exit_code == OK_PYTEST_EXIT_CODE
            get_packages_list_mock.assert_called_once_with(
                packages_dir=mock.ANY, packages_filter="all"
            )


class TestPackageTestPackagesFails(BaseAEATestCommand):
    """Test that the command 'aea test pacakages' fails with errors printed."""

    def test_packages_all_dev(self) -> None:
        """Check test packages --all fails."""

        with mock.patch("aea.cli.test.pytest.main", return_value=1), mock.patch(
            "aea.cli.test.click.echo"
        ) as echo_mock:
            result = self.run_test_command(
                "packages",
                "--all",
            )
            assert result.exit_code == 1
            assert r"call('Failed tests')" in str(echo_mock.call_args_list)
            assert r"call('1       \topen_aea/protocols/signing')" in str(
                echo_mock.call_args_list
            ).replace(r"\\\\", "/").replace(
                r"\\", "/"
            )  # fix for windows

    def test_packages_no_tests_collected(self) -> None:
        """Check test packages --all fails."""

        with mock.patch(
            "aea.cli.test.pytest.main", return_value=NO_TESTS_COLLECTED_PYTEST_EXIT_CODE
        ), mock.patch("aea.cli.test.click.echo") as echo_mock:
            result = self.run_test_command(
                "packages",
                "--all",
            )
            assert result.exit_code == 0
            assert (
                r"call('Could not collect tests for for signing of type protocols'),"
                in str(echo_mock.call_args_list)
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""A module with config tools of the aea cli."""
import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, Set

import click
import jsonschema
import yaml

from aea.cli.registry.settings import (
    DEFAULT_IPFS_URL,
    REGISTRY_CONFIG_KEY,
    REGISTRY_LOCAL,
    REMOTE_HTTP,
)
from aea.cli.utils.constants import AUTHOR_KEY, CLI_CONFIG_PATH, DEFAULT_CLI_CONFIG
from aea.cli.utils.context import Context
from aea.cli.utils.exceptions import AEAConfigException
from aea.cli.utils.generic import load_yaml
from aea.configurations.base import (
    ComponentType,
    PackageConfiguration,
    PackageType,
    _get_default_configuration_file_name_from_type,
)
from aea.configurations.constants import DEFAULT_AEA_CONFIG_FILE
from aea.configurations.loader import ConfigLoader, ConfigLoaders
from aea.configurations.validation import ExtraPropertiesError, _SCHEMAS_DIR
from aea.exceptions import AEAEnforceError, AEAValidationError
from aea.helpers.io import open_file


def try_to_load_agent_config(
    ctx: Context, is_exit_on_except: bool = True, agent_src_path: Optional[str] = None
) -> None:
    """
    Load agent config to a click context object.

    :param ctx: click command context object.
    :param is_exit_on_except: bool option to exit on exception (default = True).
    :param agent_src_path: path to an agent dir if needed to load a custom config.
    """
    if agent_src_path is None:
        agent_src_path = ctx.cwd

    try:
        path = Path(os.path.join(agent_src_path, DEFAULT_AEA_CONFIG_FILE))
        with open_file(path, mode="r", encoding="utf-8") as fp:
            ctx.agent_config = ctx.agent_loader.load(fp)
            ctx.agent_config.directory = Path(agent_src_path)
    except FileNotFoundError:
        if is_exit_on_except:
            raise click.ClickException(
                "Agent configuration file '{}' not found in the current directory.".format(
                    DEFAULT_AEA_CONFIG_FILE
                )
            )
    except (
        jsonschema.exceptions.ValidationError,
        ExtraPropertiesError,
        AEAValidationError,
    ) as e:
        if is_exit_on_except:
            raise click.ClickException(
                "Agent configuration file '{}' is invalid: `{}`. Please check the documentation.".format(
                    DEFAULT_AEA_CONFIG_FILE, str(e)
                )
            )
    except AEAEnforceError as e:
        raise click.ClickException(str(e))  # pragma: nocover


def validate_cli_config(config: Dict) -> None:
    """Validate CLI config using config schema."""
    with open_file(Path(_SCHEMAS_DIR, "cli_config.json")) as fp:
        schema = json.load(fp)

    validator = jsonschema.Draft4Validator(schema=schema)
    validator.validate(config)


def _init_cli_config(config: Optional[Dict] = None) -> None:
    """Create cli config folder and file."""
    config = config or DEFAULT_CLI_CONFIG
    conf_dir = os.path.dirname(CLI_CONFIG_PATH)
    if not os.path.exists(conf_dir):
        os.makedirs(conf_dir)
    with open_file(CLI_CONFIG_PATH, "w+") as f:
        yaml.dump(config, f, default_flow_style=False)


def update_cli_config(dict_conf: Dict) -> None:
    """
    Update CLI config and write to yaml file.

    :param dict_conf: dict config to write.
    """
    config = get_or_create_cli_config()
    config.update(dict_conf)

    validate_cli_config(config)
    _init_cli_config(config)


def get_or_create_cli_config() -> Dict:
    """
    Read or create CLI config from yaml file.

    :return: dict CLI config.
    """
    try:
        config = load_yaml(CLI_CONFIG_PATH)
    except FileNotFoundError:
        # We cannnot use _init_cli_config() anymore since it wiil inturrupt with
        # aea init command. Since we load registry info at top of the command group
        # with this function call it'll create a default ~/.aea/cli_config.yaml file
        # before reaching to the do_init function call and leading to the command
        # failing saying the config already exists.
        config = DEFAULT_CLI_CONFIG.copy()

    validate_cli_config(config)
    return config


def set_cli_author(click_context: click.Context) -> None:
    """
    Set CLI author in the CLI Context.

    The key of the new field is 'cli_author'.

    :param click_context: the Click context
    """
    config = get_or_create_cli_config()
    cli_author = config.get(AUTHOR_KEY, None)
    if cli_author is None:
        raise click.ClickException(
            "The AEA configurations are not initialized. Use `aea init` before continuing."
        )
    click_context.obj.set_config("cli_author", cli_author)


def get_registry_config() -> Dict:
    """Returns registry config from CLI config."""

    config = get_or_create_cli_config()
    cli_registry_config = config.get(REGISTRY_CONFIG_KEY, None)
    if cli_registry_config is None:
        raise click.ClickException(
            "The AEA configurations are not initialized. Use `aea init` before continuing."
        )
    return cli_registry_config


def get_registry_path_from_cli_config() -> Optional[str]:
    """Get registry path from config."""
    return (
        get_or_create_cli_config()
        .get(REGISTRY_CONFIG_KEY, {})
        .get("settings", {})
        .get(REGISTRY_LOCAL, {})
        .get("default_packages_path")
    )


def get_default_author_from_cli_config() -> Optional[str]:
    """Get registry path from config."""
    config = get_or_create_cli_config()
    return config.get(AUTHOR_KEY, None)


def get_default_remote_registry(default: str = REMOTE_HTTP) -> str:
    """Return remote registry from cli config."""
    return (
        get_or_create_cli_config()
        .get("registry_config", {})
        .get("settings", {})
        .get("remote", {})
        .get("default", default)
    )


def get_ipfs_node_multiaddr(default: str = DEFAULT_IPFS_URL) -> str:
    """Return remote registry from cli config."""
    return (
        get_or_create_cli_config()
        .get("registry_config", {})
        .get("settings", {})
        .get("remote", {})
        .get("ipfs", {})
        .get("ipfs_node", default)
    )


def load_item_config(
    item_type: str, package_path: Path, package_type_config_class: Optional[Dict] = None
) -> PackageConfiguration:
    """
    Load item configuration.

    :param item_type: type of item.
    :param package_path: path to package from which config should be loaded.
    :param package_type_config_class: Package type to config loader mappings.

    :return: configuration object.
    """
    configuration_file_name = _get_default_configuration_file_name_from_type(item_type)
    configuration_path = package_path / configuration_file_name
    configuration_loader = ConfigLoader.from_configuration_type(
        PackageType(item_type), package_type_config_class=package_type_config_class
    )
    with open_file(configuration_path) as file_input:
        item_config = configuration_loader.load(file_input)
    return item_config


def dump_item_config(
    package_configuration: PackageConfiguration,
    package_path: Path,
    package_type_config_class: Optional[Dict] = None,
) -> None:
    """
    Dump item configuration.

    :param package_configuration: the package configuration.
    :param package_path: path to package from which config should be dumped.
    :param package_type_config_class: Package type to config loader mappings.
    """
    configuration_file_name = _get_default_configuration_file_name_from_type(
        package_configuration.package_type
    )
    configuration_path = package_path / configuration_file_name
    configuration_loader = ConfigLoader.from_configuration_type(
        package_configuration.package_type,
        package_type_config_class=package_type_config_class,
    )
    with configuration_path.open("w") as file_output:
        configuration_loader.dump(package_configuration, file_output)  # type: ignore


def update_item_config(
    item_type: str,
    package_path: Path,
    package_type_config_class: Optional[Dict] = None,
    **kwargs: Any
) -> None:
    """
    Update item config and item config file.

    :param item_type: type of item.
    :param package_path: path to a package folder.
    :param package_type_config_class: Package type to config loader mappings.
    :param kwargs: pairs of config key-value to update.
    """
    item_config = load_item_config(item_type, package_path)
    for key, value in kwargs.items():
        setattr(item_config, key, value)

    config_filepath = os.path.join(
        package_path, item_config.default_configuration_filename
    )
    loader = ConfigLoaders.from_package_type(
        item_type, package_type_config_class=package_type_config_class
    )
    with open_file(config_filepath, "w") as f:
        loader.dump(item_config, f)


def validate_item_config(
    item_type: str,
    package_path: Path,
    package_type_config_class: Optional[Dict] = None,
) -> None:
    """
    Validate item configuration.

    :param item_type: type of item.
    :param package_path: path to a package folder.
    :param package_type_config_class: Package type to config loader mappings.

    :raises AEAConfigException: if something is wrong with item configuration.
    """
    item_config = load_item_config(
        item_type, package_path, package_type_config_class=package_type_config_class
    )
    loader = ConfigLoaders.from_package_type(
        item_type, package_type_config_class=package_type_config_class
    )
    for field_name in loader.required_fields:
        try:
            getattr(item_config, field_name)
        except AttributeError:
            raise AEAConfigException(
                "Parameter '{}' is missing from {} config.".format(
                    field_name, item_type
                )
            )


def get_non_vendor_package_path(aea_project_path: Path) -> Set[Path]:
    """
    Get all the paths to non-vendor packages.

    :param aea_project_path: the path to an AEA project.
    :return: the set of paths, one for each non-vendor package configuration file.
    """
    result: Set[Path] = set()
    for item_type_plural in ComponentType.plurals():
        nonvendor_package_dir_of_type = aea_project_path / item_type_plural
        result = result.union(
            {p for p in nonvendor_package_dir_of_type.iterdir() if p.is_dir()}
            if nonvendor_package_dir_of_type.exists()
            else {}
        )
    return result



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea manager."""
import contextlib
import logging
import os
import pickle  # nosec
import re
import shutil
import time
from contextlib import suppress
from copy import copy
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from typing import Callable, Optional
from unittest.case import TestCase
from unittest.mock import Mock, patch

import _strptime  # pylint: disbale=unsed-import # noqa: F401
import click
import pytest

from aea.configurations.base import PublicId
from aea.crypto.helpers import create_private_key
from aea.crypto.plugin import load_all_plugins
from aea.crypto.registries import crypto_registry
from aea.helpers.install_dependency import call_pip
from aea.manager import MultiAgentManager
from aea.manager.manager import AgentRunProcessTask, ProjectPackageConsistencyCheckError
from aea.manager.project import Project

from packages.fetchai.connections.stub.connection import StubConnection
from packages.fetchai.skills.echo import PUBLIC_ID as ECHO_SKILL_PUBLIC_ID

from tests.common.utils import wait_for_condition
from tests.conftest import (
    MAX_FLAKY_RERUNS,
    MY_FIRST_AEA_PUBLIC_ID,
    PACKAGES_DIR,
    ROOT_DIR,
)


try:
    from multiprocessing.popen_spawn_posix import _DupFd  # type: ignore # noqa: F401
except AttributeError:
    pass


pytestmark = pytest.mark.skip(
    reason="Unknown issue: https://github.com/valory-xyz/open-aea/actions/runs/7162860465/job/19500538744?pr=698"
)

DEFAULT_TIMEOUT = 120
NOT_BEFORE = "2022-01-01"
NOT_AFTER = "2023-01-01"


def default_wait_for(condition: Callable) -> None:
    """Default wait for used in tests"""
    wait_for_condition(condition, timeout=DEFAULT_TIMEOUT, period=0.5)


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class BaseCase(TestCase):
    """Base case setup/teardown."""

    MODE = "async"
    PASSWORD: Optional[str] = None

    echo_skill_id = ECHO_SKILL_PUBLIC_ID

    def setUp(self):
        """Set test case."""
        self.agent_name = "test_what_ever12"
        self.project_public_id = MY_FIRST_AEA_PUBLIC_ID
        self.tmp_dir = TemporaryDirectory()
        self.working_dir = os.path.join(self.tmp_dir.name, "MultiAgentManager_dir")
        self.project_path = os.path.join(
            self.working_dir, self.project_public_id.author, self.project_public_id.name
        )
        assert not os.path.exists(self.working_dir)
        self.manager = MultiAgentManager(
            self.working_dir, mode=self.MODE, password=self.PASSWORD
        )
        self.log_file = f"{self.tmp_dir.name}/{self.agent_name}.log"
        self._log_handlers = copy(logging.root.getChild("aea").handlers)

    def tearDown(self):
        """Tear down test case."""
        try:
            self.manager.stop_manager()

            for task in self.manager._agents_tasks.values():
                if isinstance(task, AgentRunProcessTask):
                    task.process.terminate()
                    task.process.join(AgentRunProcessTask.PROCESS_JOIN_TIMEOUT)

            time.sleep(1)
            logging.shutdown(
                [
                    handler
                    for handler in logging._handlerList
                    if getattr(handler, "baseFilename", None)
                ]
            )
            logging.root.getChild("aea").handlers = [
                handler
                for handler in self._log_handlers
                if not getattr(handler, "baseFilename", None)
            ]
            time.sleep(1)
            if os.path.exists(self.working_dir):
                rmtree(self.working_dir)
        finally:
            self.tmp_dir.cleanup()


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class TestMultiAgentManagerDependencies(BaseCase):
    """Test plugin installed and loaded as a dependency."""

    @pytest.mark.skip  # issue with plugins
    def test_plugin_dependencies(self, *args):
        """Test plugin installed and loaded as a dependency."""
        plugin_path = str(Path(ROOT_DIR) / "plugins" / "aea-ledger-fetchai")
        install_cmd = f"install --no-deps {plugin_path}".split(" ")
        try:
            self.manager.start_manager()
            call_pip("uninstall open-aea-ledger-fetchai -y".split(" "))
            from aea.crypto.registries import ledger_apis_registry

            ledger_apis_registry.specs.pop("fetchai", None)
            load_all_plugins(is_raising_exception=False)
            assert "fetchai" not in ledger_apis_registry.specs

            self.manager.add_project(self.project_public_id, local=True)
            assert "fetchai" not in ledger_apis_registry.specs

            self.manager.remove_project(self.project_public_id)

            def install_deps(*_):
                call_pip(install_cmd)

            with patch(
                "aea.aea_builder.AEABuilder.install_pypi_dependencies", install_deps
            ):
                self.manager.add_project(self.project_public_id, local=True)

            assert "fetchai" in ledger_apis_registry.specs
        finally:
            call_pip("uninstall open-aea-ledger-fetchai -y".split(" "))
            call_pip(install_cmd)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
@patch("aea.aea_builder.AEABuilder.check_project_dependencies")
class BaseTestMultiAgentManager(BaseCase):
    """Base test class for multi-agent manager"""

    def test_workdir_created_removed(self, *args):
        """Check work dit created removed on MultiAgentManager start and stop."""
        assert not os.path.exists(self.working_dir)
        self.manager.start_manager()
        assert os.path.exists(self.working_dir)
        self.manager.stop_manager()
        assert not os.path.exists(self.working_dir)
        assert not os.path.exists(self.working_dir)

    def test_projects_property(self, *args):
        """Test projects property."""
        self.assertEqual(self.manager.projects, self.manager._projects)

    def test_data_dir_presents(self, *args):
        """Check not fails on exists data dir."""
        try:
            os.makedirs(self.working_dir)
            os.makedirs(self.manager._data_dir)
            self.manager.start_manager()
            self.manager.stop_manager()
        finally:
            with suppress(Exception):
                rmtree(self.working_dir)

    def test_MultiAgentManager_is_running(self, *args):
        """Check MultiAgentManager is running property reflects state."""
        assert not self.manager.is_running
        self.manager.start_manager()
        assert self.manager.is_running
        self.manager.stop_manager()
        assert not self.manager.is_running

    def test_add_remove_project(self, *args):
        """Test add and remove project."""
        self.manager.start_manager()

        self.manager.add_project(self.project_public_id, local=True)

        assert self.project_public_id in self.manager.list_projects()
        assert os.path.exists(self.project_path)

        with pytest.raises(ValueError, match=r".*was already added.*"):
            self.manager.add_project(self.project_public_id, local=True)

        self.manager.remove_project(self.project_public_id)
        assert self.project_public_id not in self.manager.list_projects()

        with pytest.raises(ValueError, match=r"is not present"):
            self.manager.remove_project(self.project_public_id)

        self.manager.add_project(self.project_public_id, local=True)
        assert self.project_public_id in self.manager.list_projects()
        assert os.path.exists(self.project_path)

    def test_add_project_check_version_consistency(self, *args):
        """Test add project check version consistency"""

        self.manager.start_manager()

        error_msg = "The remote registry is not initialized. Remote registry command currently not supported - manually edit config file!"
        with pytest.raises(click.ClickException, match=error_msg):
            self.manager.add_project(self.project_public_id, remote=True)

        self.manager.add_project(self.project_public_id)
        project = self.manager._projects[self.project_public_id]
        prefix_to_version = self.manager._package_id_prefix_to_version
        package_prefix, (version, agents) = prefix_to_version.copy().popitem()

        with patch.object(
            self.manager, "_versionless_projects_set", return_value=set()
        ):
            with patch.object(Project, "load", return_value=project):
                self.manager.add_project(self.project_public_id)
                prefix_to_version[package_prefix] = version + "0", agents
                with pytest.raises(
                    ProjectPackageConsistencyCheckError,
                    match="AEA dependencies have conflicts with previously added projects",
                ):
                    self.manager.add_project(self.project_public_id)

    def read_logs(self) -> str:
        """Read log file"""
        return open(self.log_file, "r").read()

    def test_add_agent(self, *args):
        """Test add agent alias."""
        self.manager.start_manager()

        self.manager.add_project(self.project_public_id, local=True)

        new_tick_interval = 0.2111

        component_overrides = [
            {
                **self.echo_skill_id.json,
                "type": "skill",
                "behaviours": {"echo": {"args": {"tick_interval": new_tick_interval}}},
            }
        ]
        agent_overrides = {
            "logging_config": {
                "version": 1,
                "formatters": {"standard": {"format": ""}},
                "handlers": {
                    "console": {
                        "class": "logging.StreamHandler",
                        "formatter": "standard",
                        "level": "DEBUG",
                    },
                    "file": {
                        "class": "logging.FileHandler",
                        "filename": self.log_file,
                        "mode": "w",
                        "level": "DEBUG",
                        "formatter": "standard",
                    },
                },
                "loggers": {"aea": {"level": "DEBUG", "handlers": ["file", "console"]}},
            }
        }
        self.manager.add_agent(
            self.project_public_id,
            self.agent_name,
            component_overrides=component_overrides,
            agent_overrides=agent_overrides,
        )

        agent_alias = self.manager.get_agent_alias(self.agent_name)
        assert agent_alias.agent_name == self.agent_name
        assert (
            agent_alias.get_aea_instance()
            .resources.get_behaviour(self.echo_skill_id, "echo")
            .tick_interval
            == new_tick_interval
        )

        with pytest.raises(ValueError, match="already exists"):
            self.manager.add_agent(
                self.project_public_id,
                self.agent_name,
            )

    def test_set_overrides(self, *args):
        """Test agent set overrides."""
        self.test_add_agent()
        new_tick_interval = 1000.0
        component_overrides = [
            {
                **self.echo_skill_id.json,
                "type": "skill",
                "behaviours": {"echo": {"args": {"tick_interval": new_tick_interval}}},
            }
        ]
        self.manager.set_agent_overrides(
            self.agent_name,
            agent_overides=None,
            components_overrides=component_overrides,
        )
        agent_alias = self.manager.get_agent_alias(self.agent_name)
        assert agent_alias.agent_name == self.agent_name
        assert (
            agent_alias.get_aea_instance()
            .resources.get_behaviour(self.echo_skill_id, "echo")
            .tick_interval
            == new_tick_interval
        )

    def test_remove_agent(self, *args):
        """Test remove agent alias."""
        self.test_add_agent()
        assert self.agent_name in self.manager.list_agents()
        self.manager.remove_agent(self.agent_name)
        assert self.agent_name not in self.manager.list_agents()

        with pytest.raises(ValueError, match="does not exist!"):
            self.manager.remove_agent(self.agent_name)

    def test_remove_project_with_alias(self, *args):
        """Test remove project with alias presents."""
        self.test_add_agent()

        with pytest.raises(
            ValueError, match="Can not remove projects with aliases exists"
        ):
            self.manager.remove_project(self.project_public_id)

    def test_add_agent_for_non_exist_project(self, *args):
        """Test add agent when no project added."""
        with pytest.raises(ValueError, match=" project is not added"):
            self.manager.add_agent(PublicId("test", "test", "0.1.0"), "another_agent")

    def test_agent_actually_running(self, *args):
        """Test MultiAgentManager starts agent correctly and agent perform acts."""
        self.test_add_agent()

        self.manager.start_all_agents()

        default_wait_for(
            lambda: "Echo Behaviour: act method called." in self.read_logs()
        )

        with contextlib.suppress(Exception):
            self.manager.stop_agent(self.agent_name)

    def test_exception_handling(self, *args):
        """Test error callback works."""
        self.add_agent("test_agent", PublicId.from_str("fetchai/error_test:0.1.0"))
        self.manager.start_all_agents()

        callback_mock = Mock()

        self.manager.add_error_callback(callback_mock)

        self.manager.start_all_agents()
        default_wait_for(lambda: callback_mock.call_count > 0)

    def add_agent(self, agent_name: str, project_id: PublicId) -> None:
        """Add agent and start manager."""
        self.manager.start_manager()

        self.manager.add_project(project_id, local=True)

        self.manager.add_agent(
            project_id,
            agent_name,
        )

        agent_alias = self.manager.get_agent_alias(agent_name)
        assert agent_alias

    def test_default_exception_handling(self, *args):
        """Test that the default error callback works."""
        self.add_agent("test_agent", PublicId.from_str("fetchai/error_test:0.1.0"))

        with patch.object(
            self.manager,
            "_print_exception_occurred_but_no_error_callback",
            side_effect=self.manager._print_exception_occurred_but_no_error_callback,
        ) as callback_mock:
            self.manager.start_all_agents()
            default_wait_for(lambda: callback_mock.call_count > 0)
            callback_mock.assert_called_once()

    def test_stop_from_exception_handling(self, *args):
        """Test stop MultiAgentManager from error callback."""
        self.add_agent("test_agent", PublicId.from_str("fetchai/error_test:0.1.0"))

        def handler(*args, **kwargs):
            self.manager.stop_manager()

        self.manager.add_error_callback(handler)

        self.manager.start_all_agents()
        default_wait_for(lambda: not self.manager.is_running)

    def test_start_all(self, *args):
        """Test MultiAgentManager start all agents."""
        self.test_add_agent()
        assert self.agent_name in self.manager.list_agents()
        assert self.agent_name not in self.manager.list_agents(running_only=True)
        self.manager.start_all_agents()
        assert self.agent_name in self.manager.list_agents(running_only=True)

        self.manager.start_all_agents()

        # check every agent started
        default_wait_for(
            lambda: len(self.manager.list_agents())
            == len(self.manager.list_agents(running_only=True))
        )

        default_wait_for(lambda: "Start processing messages" in self.read_logs())

        with pytest.raises(ValueError, match="is already started!"):
            self.manager.start_agents(self.manager.list_agents())

        with pytest.raises(ValueError, match="is already started!"):
            self.manager.start_agent(self.agent_name)

        with pytest.raises(ValueError, match="is not registered!"):
            self.manager.start_agent("non_exists_agent")

    def test_stop_agent(self, *args):
        """Test stop agent."""
        self.test_start_all()
        default_wait_for(lambda: self.manager.list_agents(running_only=True))

        default_wait_for(lambda: "Start processing messages" in self.read_logs())

        self.manager.stop_all_agents()

        default_wait_for(lambda: not self.manager.list_agents(running_only=True))

        assert not self.manager.list_agents(running_only=True)

        with pytest.raises(ValueError, match=" is not running!"):
            self.manager.stop_agent(self.agent_name)

        with pytest.raises(ValueError, match=" is not running!"):
            self.manager.stop_agents([self.agent_name])

        default_wait_for(lambda: "Runtime loop stopped!" in self.read_logs())

    def test_do_no_allow_override_some_fields(self, *args):
        """Do not allo to override some values in agent config."""
        self.manager.start_manager()

        self.manager.add_project(self.project_public_id, local=True)

        BAD_OVERRIDES = [
            "skills",
            "connections",
            "contracts",
            "protocols",
            "some_field?",
        ]

        for bad_override in BAD_OVERRIDES:
            with pytest.raises(
                ValueError, match=r"Attribute `.*` is not allowed to be updated!"
            ):
                self.manager.add_agent(
                    self.project_public_id,
                    self.agent_name,
                    agent_overrides={bad_override: "some value"},
                )

    @staticmethod
    def test_invalid_mode(*args):
        """Test MultiAgentManager fails on invalid mode."""
        with pytest.raises(ValueError, match="Invalid mode"):
            MultiAgentManager("test_dir", mode="invalid_mode")

    def test_double_start(self, *args):
        """Test double MultiAgentManager start."""
        self.manager.start_manager()
        assert self.manager.is_running
        self.manager.start_manager()
        assert self.manager.is_running

    def test_double_stop(self, *args):
        """Test double MultiAgentManager stop."""
        self.manager.start_manager()
        assert self.manager.is_running
        self.manager.stop_manager()
        assert not self.manager.is_running
        self.manager.stop_manager()
        assert not self.manager.is_running

    def test_remove_running_agent(self, *args):
        """Test fail on remove running agent."""
        self.test_start_all()

        default_wait_for(lambda: "Start processing messages" in self.read_logs())

        with pytest.raises(ValueError, match="Agent is running. stop it first!"):
            self.manager.remove_agent(self.agent_name)

        self.manager.stop_all_agents()

        # wait all stopped
        default_wait_for(lambda: not self.manager.list_agents(running_only=True))
        assert self.agent_name not in self.manager.list_agents(running_only=True)

        self.manager.remove_agent(self.agent_name)
        assert self.agent_name not in self.manager.list_agents()

    def test_save_load_positive(self, *args):
        """Test save-load func of MultiAgentManager for positive result."""
        self.manager.start_manager()
        self.manager.add_project(self.project_public_id, local=True)

        self.manager.add_agent(self.project_public_id, self.agent_name)
        self.manager.stop_manager(save=True)
        assert os.path.exists(self.manager._save_path)

        self.manager.start_manager()
        assert self.project_public_id in self.manager._projects.keys()
        assert self.agent_name in self.manager._agents.keys()

    def test_list_agents_info_positive(self, *args):
        """Test list_agents_info method for positive result."""
        self.manager.start_manager()
        self.manager.add_project(self.project_public_id, local=True)

        self.manager.add_agent(self.project_public_id, self.agent_name)
        result = self.manager.list_agents_info()
        expected_result = [
            {
                "agent_name": self.agent_name,
                "public_id": str(self.project_public_id),
                "addresses": self.manager.get_agent_alias(
                    self.agent_name
                ).get_addresses(),
                "is_running": False,
            }
        ]
        assert result == expected_result

    def test_add_same_project_versions(self, *args):
        """Test add the same project twice."""
        self.manager.start_manager()

        self.manager.add_project(self.project_public_id, local=True)
        with pytest.raises(
            ValueError, match=r"The project \(fetchai/my_first_aea\) was already added!"
        ):
            self.manager.add_project(
                PublicId.from_str("fetchai/my_first_aea:0.15.0"), remote=True
            )

    def test_get_overridables(self, *args):
        """Test get overridables."""
        self.manager.start_manager()
        self.manager.add_project(self.project_public_id, local=True)
        self.manager.add_agent(self.project_public_id, self.agent_name)

        (
            agent_overridables,
            components_overridables,
        ) = self.manager.get_agent_overridables(self.agent_name)
        assert "default_ledger" in agent_overridables
        assert "timeout" in agent_overridables
        assert "description" in agent_overridables
        assert len(components_overridables) == 2
        assert "is_abstract" in components_overridables[0]

    def test_issue_certificates(self, *args):
        """Test agent alias issue certificates."""
        self.manager.start_manager()
        self.manager.add_project(self.project_public_id, local=True)

        cert_filename = "cert.txt"
        cert_path = os.path.join(self.manager.data_dir, self.agent_name, cert_filename)
        assert not os.path.exists(cert_filename)

        priv_key_path = os.path.abspath(os.path.join(self.working_dir, "priv_key.txt"))
        create_private_key("fetchai", priv_key_path, password=self.PASSWORD)
        assert os.path.exists(priv_key_path)

        component_overrides = [
            {
                **StubConnection.connection_id.json,
                "type": "connection",
                "cert_requests": [
                    {
                        "identifier": "acn",
                        "ledger_id": "fetchai",
                        "not_before": NOT_BEFORE,
                        "not_after": NOT_AFTER,
                        "public_key": "fetchai",
                        "message_format": "{public_key}",
                        "save_path": cert_filename,
                    }
                ],
            }
        ]

        agent_overrides = {
            "private_key_paths": {"fetchai": priv_key_path},
            "connection_private_key_paths": {"fetchai": priv_key_path},
        }
        self.manager.add_agent(
            self.project_public_id,
            self.agent_name,
            agent_overrides=agent_overrides,
            component_overrides=component_overrides,
        )
        agent_alias = self.manager.get_agent_alias(self.agent_name)

        agent_alias.issue_certificates()

        assert os.path.exists(cert_path)

    def test_get_addresses(self, *args) -> None:
        """Test get addresses for agent alias."""
        self.test_add_agent()
        agent_alias = self.manager.get_agent_alias(self.agent_name)
        keys = {
            name: agent_alias._create_private_key(
                ledger=name, replace=True, is_connection=False
            )
            for name in crypto_registry.supported_ids
        }

        connection_keys = {
            name: agent_alias._create_private_key(
                ledger=name, replace=True, is_connection=True
            )
            for name in crypto_registry.supported_ids
        }
        agent_alias.set_overrides(
            {"private_key_paths": keys, "connection_private_key_paths": connection_keys}
        )

        assert len(agent_alias.get_addresses()) == len(crypto_registry.supported_ids)
        assert len(agent_alias.get_connections_addresses()) == len(
            crypto_registry.supported_ids
        )

    def test_addresses_autoadded(self, *args) -> None:
        """Test addresses automatically added on creation."""
        self.test_add_agent()
        agent_alias = self.manager.get_agent_alias(self.agent_name)
        assert len(agent_alias.get_addresses()) == 1
        assert len(agent_alias.get_connections_addresses()) == 1


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class TestMultiAgentManagerAsyncMode(
    BaseTestMultiAgentManager
):  # pylint: disable=unused-argument,protected-access,attribute-defined-outside-init
    """Tests for MultiAgentManager in async mode."""


class TestMultiAgentManagerAsyncModeWithPassword(
    BaseTestMultiAgentManager
):  # pylint: disable=unused-argument,protected-access,attribute-defined-outside-init
    """Tests for MultiAgentManager in async mode, with password."""

    PASSWORD = "password"  # nosec


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class TestMultiAgentManagerThreadedMode(BaseTestMultiAgentManager):
    """Tests for MultiAgentManager in threaded mode."""

    MODE = "threaded"


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class TestMultiAgentManagerMultiprocessMode(BaseTestMultiAgentManager):
    """Tests for MultiAgentManager in multiprocess mode."""

    MODE = "multiprocess"

    def test_plugin_dependencies(self, *args):
        """Skip test cause multiprocess works another way."""


class TestMultiAgentManagerMultiprocessModeWithPassword(
    TestMultiAgentManagerMultiprocessMode
):
    """Tests for MultiAgentManager in multiprocess mode with password."""

    PASSWORD = "password"  # nosec


class TestMultiAgentManagerThreadedModeWithPassword(BaseTestMultiAgentManager):
    """Tests for MultiAgentManager in threaded mode, with password."""

    MODE = "threaded"
    PASSWORD = "password"  # nosec


@patch("aea.aea_builder.AEABuilder.install_pypi_dependencies")
class TestMultiAgentManagerPackageConsistencyError:
    """
    Test that the MultiAgentManager (MAM) raises an error on package version inconsistency.

    In particular, an invariant to keep for the MAM state is that
    there are no two AEA package dependencies across all the projects
    that have different versions. This is due to a known limitation
    in the package loading system.

    This test checks that when the invariant is violated when
    the method "add_project" is called, the framework raises an
    exception with a descriptive error message.
    """

    EXPECTED_ERROR_MESSAGE = """cannot add project 'open_aea/my_first_aea:0.1.0': the following AEA dependencies have conflicts with previously added projects"""

    def setup(self):
        """Set up the test case."""
        self.project_public_id = MY_FIRST_AEA_PUBLIC_ID
        self.tmp_dir = TemporaryDirectory()
        self.manager = MultiAgentManager(self.tmp_dir.name)

    def test_run(self, *args):
        """
        Run the test.

        First add agent project "open_aea/my_first_aea:0.1.0".
        Then we mutate the stored `_package_id_prefix_to_version`
        change the path and mock the `_versionless_projects_set`.
        and attempt to add it anew, as if a different project.
        The second addition will fail because the projects
        contain conflicting AEA package versions.
        """
        self.manager.start_manager()
        my_first_aea_id = PublicId.from_str("open_aea/my_first_aea:0.1.0")
        self.manager.add_project(my_first_aea_id)
        base = Path(self.tmp_dir.name, "open_aea")
        shutil.move(base / "my_first_aea", base / "my_first_aea_bumped")
        p2v = self.manager._package_id_prefix_to_version
        for k, (_, public_id_set) in p2v.items():
            p2v[k] = ("0.0.0", public_id_set)
        with pytest.raises(
            ProjectPackageConsistencyCheckError,
            match=re.escape(self.EXPECTED_ERROR_MESSAGE),
        ):
            with patch.object(self.manager, "_versionless_projects_set"):
                self.manager.add_project(my_first_aea_id)

    def teardown(self):
        """Tear down test case."""
        try:
            self.manager.stop_manager()
        finally:
            self.tmp_dir.cleanup()


class TestMultiAgentManagerWithPotentiallyConflictingPackages:
    """
    Test that the MultiAgentManager (MAM) handles correctly potentially conflicting packages.

    "Potentially conflicting packages" are packages that have the same
    package id prefix. The reason why they are "potentially conflicting"
    is that their version may be different, i.e. conflicting,
    and therefore are incompatible within the same MAM.

    This test checks that, when adding a new project,
    in case there are potentially conflicting packages but that
    are not in fact conflicting, the operation is completed successfully.
    """

    def setup(self):
        """Set up the test case."""
        self.project_public_id = MY_FIRST_AEA_PUBLIC_ID
        self.tmp_dir = TemporaryDirectory()
        self.manager = MultiAgentManager(self.tmp_dir.name)

    def test_run(self):
        """
        Run the test.

        First add agent project "open_aea/my_first_aea:0.1.0".
        Then we change the path and mock the `_versionless_projects_set`,
        and attempt to add it anew, as if a different project.
        The second addition will fail because the projects
        contain conflicting AEA package versions.
        """
        self.manager.start_manager()
        my_first_aea_id = PublicId.from_str("open_aea/my_first_aea:0.1.0")
        self.manager.add_project(my_first_aea_id)
        base = Path(self.tmp_dir.name, "open_aea")
        shutil.move(base / "my_first_aea", base / "my_first_aea_bumped")
        with patch.object(self.manager, "_versionless_projects_set"):
            self.manager.add_project(my_first_aea_id)

    def teardown(self):
        """Tear down test case."""
        try:
            self.manager.stop_manager()
        finally:
            self.tmp_dir.cleanup()


def test_project_auto_added_removed():
    """Check project auto added and auto removed on agent added/removed."""
    agent_name = "test_agent"
    with TemporaryDirectory() as tmp_dir, patch(
        "aea.manager.project.Project.build"
    ), patch("aea.manager.project.Project.install_pypi_dependencies"):
        manager = MultiAgentManager(
            tmp_dir,
            mode="async",
            registry_path=PACKAGES_DIR,
            auto_add_remove_project=True,
        )
        try:
            manager.start_manager()
            assert not manager.list_projects()
            manager.add_agent(
                PublicId("fetchai", "my_first_aea"), agent_name, local=True
            )
            assert manager.list_projects()
            assert manager.list_agents()
            manager.remove_agent(agent_name)
            assert not manager.list_agents()
            assert not manager.list_projects()
        finally:
            manager.stop_manager()


def test_dump():
    """Check project auto added and auto removed on agent added/removed."""
    agent_name = "test_agent"
    with TemporaryDirectory() as tmp_dir, patch(
        "aea.manager.project.Project.build"
    ), patch("aea.manager.project.Project.install_pypi_dependencies"):
        manager = MultiAgentManager(
            tmp_dir,
            mode="async",
            registry_path=PACKAGES_DIR,
            auto_add_remove_project=True,
        )
        try:
            manager.start_manager()
            assert not manager.list_projects()
            manager.add_agent(
                PublicId("fetchai", "my_first_aea"), agent_name, local=True
            )
            alias = manager.get_agent_alias(agent_name)
            data = pickle.dumps(alias)
            alias2 = pickle.loads(data)  # nosec
            assert alias.dict == alias2.dict
            assert alias.project.public_id == alias2.project.public_id

            assert manager.list_projects()
            assert manager.list_agents()
            manager.remove_agent(agent_name)
            assert not manager.list_agents()
            assert not manager.list_projects()
        finally:
            manager.stop_manager()


def test_handle_error_on_load_state():
    """Check project auto added and auto removed on agent added/removed."""
    agent_name = "test_agent"
    with TemporaryDirectory() as tmp_dir:
        manager = MultiAgentManager(
            tmp_dir,
            mode="async",
            registry_path=PACKAGES_DIR,
            auto_add_remove_project=True,
        )

        with pytest.raises(ValueError, match="Manager was not started"):
            manager.last_start_status

        try:
            manager.start_manager()
            state_loaded, *_ = manager.last_start_status
            assert not state_loaded
            assert not manager.list_projects()
            manager.add_agent(
                PublicId("fetchai", "my_first_aea"), agent_name, local=True
            )
            assert manager.list_projects()
            manager._save_state()
        finally:
            manager.stop_manager(cleanup=False)

        # check loaded well
        manager = MultiAgentManager(
            tmp_dir,
            mode="async",
            registry_path=PACKAGES_DIR,
            auto_add_remove_project=False,
        )
        try:
            manager.start_manager()
            state_loaded, loaded_ok, *_ = manager.last_start_status
            assert state_loaded
            assert len(loaded_ok) == 1
            assert manager.list_projects()
            assert manager.list_agents()
        finally:
            manager.stop_manager(cleanup=False)

        config_file = (
            Path(tmp_dir) / "fetchai/my_first_aea/vendor/fetchai/skills/echo/skill.yaml"
        )
        config_yaml = config_file.read_text()
        new_version = "'>=0.0.1, <0.0.2'"
        new_config = re.sub(
            r"'>=[0-9]+.[0-9]+.[0-9]+, <[0-9]+.[0-9]+.[0-9]+'",
            new_version,
            config_yaml,
        )
        assert new_version in new_config
        config_file.write_text(new_config)
        # check load fails
        manager = MultiAgentManager(
            tmp_dir,
            mode="async",
            registry_path=PACKAGES_DIR,
            auto_add_remove_project=False,
        )
        try:
            manager.start_manager()
            state_loaded, loaded_ok, load_failed = manager.last_start_status
            assert state_loaded
            assert len(loaded_ok) == 0
            assert len(load_failed) == 1
            assert isinstance(load_failed[0][0], PublicId)
            assert isinstance(load_failed[0][1], list)
            assert len(load_failed[0][1]) == 1
            assert isinstance(load_failed[0][1][0], dict)
            assert isinstance(load_failed[0][2], Exception)
            assert re.match(
                "Failed to load project: fetchai/my_first_aea:latest Error: The CLI version is .*, but package fetchai/echo:0.19.0 requires version <0.0.2,>=0.0.1",
                str(load_failed[0][2]),
            )
            assert not manager.list_projects()
            assert not manager.list_agents()
        finally:
            manager.stop_manager()
